# Qt imports
from PySide2.QtWidgets import QWidget
from PySide2.QtCore import Signal

# Custom imports
from cutevariant.core import Query


class PluginWidget(QWidget):
    """Handy class for common methods of QueryPluginWidget and VariantPluginWidget"""

    def __init__(self, parent=None):
        super().__init__(parent)

    def objectName(self):
        """Override: Return an object name based on windowTitle

        .. note:: Some plugins don't set objectName attribute and so their state
            can't be saved with MainWindow's saveState function.
        """
        return self.windowTitle().lower()


class QueryPluginWidget(PluginWidget):
    """Base class for all widgets which observe a Query

    .. seealso:: queryrooter.py
    """

    # Signals
    # When the query's widget changed
    query_changed = Signal()
    # When the widget emits a message to be displayed in the status bar
    message = Signal(str)

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, query: Query):
        self._query = query
        self.on_init_query()

    def on_change_query(self):
        """Called by the queryrouter each time a belong widget send a
        query_changed signal
        """
        raise NotImplementedError(self.__class__.__name__)

    def on_init_query(self):
        """Called by the queryrouter when query is set"""
        raise NotImplementedError(self.__class__.__name__)


class VariantPluginWidget(PluginWidget):
    """Base class for all widgets which get variant data when clicking on it
    from the main view
    """

    def set_variant(self, variant):
        raise NotImplementedError(self.__class__.__name__)



"""Proof of concept to display the 3D structure of a protein in a WebGL accelerated widget"""
# Qt imports
from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWebEngineWidgets import QWebEngineView

# Custom imports
from .plugin import QueryPluginWidget
from cutevariant.core import Query


class WebGLQueryWidget(QueryPluginWidget):
    """Display the 3D structure of a protein in a WebGL accelerated widget"""

    def __init__(self, protein_reference="1crn", parent=None):
        """Display the 3D structure of the given protein"""
        super().__init__(parent)
        self.setWindowTitle("web GL")

        self.view = QWebEngineView()
        # self.view.loadFinished.connect(self.loaded)
        self.view.setHtml(
            """
            <!DOCTYPE html>
            <html lang="en">
            <head>
              <meta charset="utf-8">
              <script src="https://cdn.rawgit.com/arose/ngl/v2.0.0-dev.31/dist/ngl.js"></script>
            </head>
            <body>

            <style type="text/css">
            * { margin: 0; padding: 0; }
            html, body { width: 100%; height: 100%; overflow: hidden; }
            </style>

            <div id="viewport" style="width:100%; height:100%;"></div>

            <script>
              // Create NGL Stage object
              var stage = new NGL.Stage( "viewport" );

              // Handle window resizing
              window.addEventListener( "resize", function( event ){
                stage.handleResize();
              }, false );

              // Load PDB entry
              stage.loadFile( "rcsb://6J3G", { defaultRepresentation: true } );
            </script>
            </body>
            </html>
            """
            
        )

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.view)

        self.setLayout(layout)
        self._query = None

    @property
    def query(self):
        return self._query  # Useless , this widget is query read only

    @query.setter
    def query(self, query: Query):
        self._query = query

    def loaded(self):
        print("FINISHED!!!!!!!!!!!!!!!!!!!")



"""Plugin to show all characteristics of a selected variant

InfoVariantWidget is showed on the GUI, it uses VariantPopupMenu to display a
contextual menu about the variant which is selected.
VariantPopupMenu is also used in viewquerywidget for the same purpose.
"""
# Standard imports
from functools import partial

# Qt imports
from PySide2.QtCore import Qt, QPoint, QSettings, QUrl
from PySide2.QtWidgets import *
from PySide2.QtGui import QDesktopServices

# Custom imports
from .plugin import VariantPluginWidget
from cutevariant.gui.settings import SettingsWidget
from cutevariant.gui.ficon import FIcon
from cutevariant.gui.style import TYPE_COLORS
from cutevariant.commons import logger, WEBSITES_URLS

LOGGER = logger()


class VariantPopupMenu(QMenu):
    """Class to display a contextual menu about the variant which is above the mouse

    .. note:: popup() is overrided and takes the variant as argument.
    """

    def __init__(self):
        super().__init__()

        self._variant = None

        # Fill menu
        self.addAction(FIcon(0xF4CE), self.tr("Add to Favorite")).setCheckable(True)
        self.addAction(
            FIcon(0xF18F), self.tr("Copy genomic location"), self.to_clipboard
        )

        # Create submenu for websites shortcuts
        self.sub_menu = self.addMenu(self.tr("Open With"))

    def sub_menu_setup(self):
        """Setup sub menu 'open with'
        .. note:: This function is called each time a context menu is requested
            in the aim of viewing new added databases into settings.
        """

        def add_action(site, url_template):
            """Build action, and connect it to a dynamically generated slot"""
            # This is not ok, Why ?
            # self.__dict__[f"open_{site}_url"] = Slot()(partial(self.open_url, url_template))

            # self.addAction(
            #     self.tr(f"Search the variant on {site}"), self, SLOT(f"open_{site}_url()")
            # )

            # Method to set the slot as a instance method
            # (if we want to use it elsewhere)
            # self.__dict__[f"open_{site}_url"] = partial(self.open_url, url_template)

            self.sub_menu.addAction(self.tr(site), partial(self.open_url, url_template))

        self.sub_menu.clear()

        # Built-in urls
        [add_action(*item) for item in WEBSITES_URLS.items()]

        # User urls - Get all child keys of the group databases_urls
        settings = QSettings()
        settings.beginGroup("databases_urls/")
        # Build menu actions
        [add_action(site, settings.value(site)) for site in settings.childKeys()]
        settings.endGroup()

        self.sub_menu.addSeparator()
        self.sub_menu.addAction(self.tr("Edit ..."), self.show_settings)

    def open_url(self, url_template):
        """Open the url based on the current variant and the given url template

        .. note:: URL ex: https://varsome.com/variant/hg19/chr17-7578406-C-A
        """

        url = url_template.format(**self._variant)

        LOGGER.info("InfoVariantWidget:open_url:: Open <%s>" % url)
        QDesktopServices.openUrl(QUrl(url))

    def to_clipboard(self):
        """Copy the current variant reference to the clipboard"""
        QApplication.clipboard().setText(
            "{chr}-{pos}-{ref}-{alt}".format(**self._variant)
        )

    def popup(self, variant: str, pos):
        """Overrided: Show the popup menu

        :param variant: Dictionnary that defines a variant (chr, pos, ref, alt).
        """
        self.sub_menu_setup()
        self._variant = variant
        super().popup(pos)

    def show_settings(self):
        """Slot to show settings window opened on the Variants page"""
        widget = SettingsWidget()
        # Open Settings on the Variants widget
        # We can also iterate over stacked widgets to find the good class name
        # w = widget.stack_widget.widget(2)
        # print(w.__class__.__name__)
        # widget.list_widget.setCurrentRow(2)
        w = widget.list_widget.findItems("Variants", Qt.MatchContains)[0]
        widget.list_widget.setCurrentItem(w)
        widget.exec()


class InfoVariantWidget(VariantPluginWidget):
    """Plugin to show all characteristics of a selected variant"""

    def __init__(self):
        super().__init__()

        self.setWindowTitle(self.tr("Info variants"))

        # Set columns of TreeWidget
        self.view = QTreeWidget()
        self.view.setColumnCount(2)
        # Set title of columns
        self.view.setHeaderLabels([self.tr("Attributes"), self.tr("Values")])

        v_layout = QVBoxLayout()
        v_layout.setContentsMargins(0, 0, 0, 0)
        v_layout.addWidget(self.view)
        self.setLayout(v_layout)

        # Create menu
        self.context_menu = VariantPopupMenu()
        # Ability to trigger the menu
        self.view.setContextMenuPolicy(Qt.CustomContextMenu)
        self.view.customContextMenuRequested.connect(self.show_menu)

        self._variant = dict()

    def populate(self):
        """Show the current variant attributes on the TreeWidget"""
        # Reset
        self.view.clear()

        # Filter None values
        g = ((key, val) for key, val in self._variant.items() if val)

        for key, val in g:
            item = QTreeWidgetItem()
            item.setText(0, str(key))
            item.setText(1, str(val))
            # item.setFlags(Qt.ItemIsSelectable|Qt.ItemIsEnabled)
            # map value type to color
            item.setIcon(0, FIcon(0xF70A, TYPE_COLORS[val.__class__.__name__]))

            self.view.addTopLevelItem(item)

    def set_variant(self, variant: dict):
        """Register and show the given variant
        Called when a user clicks on a vriant on the ViewQueryWidget"""
        self._variant = variant
        self.populate()

    def show_menu(self, pos: QPoint):
        """Show context menu associated to the current variant"""
        if not self._variant:
            return
        self.context_menu.popup(self._variant, self.view.mapToGlobal(pos))



"""Main window of Cutevariant"""
# Standard imports
import os

# Qt imports
from PySide2.QtCore import Qt, QSettings, QByteArray, QDir
from PySide2.QtWidgets import *
from PySide2.QtGui import QIcon, QKeySequence

# Custom imports
from cutevariant.core import Query, get_sql_connexion
from cutevariant.gui.ficon import FIcon
from cutevariant.gui.plugin import VariantPluginWidget, QueryPluginWidget
from cutevariant.gui.wizards import ProjectWizard
from cutevariant.gui.settings import SettingsWidget
from cutevariant.gui.viewquerywidget import ViewQueryWidget
from cutevariant.gui.columnquerywidget import ColumnQueryWidget
from cutevariant.gui.filterquerywidget import FilterQueryWidget
from cutevariant.gui.selectionquerywidget import SelectionQueryWidget
from cutevariant.gui.vqleditor import VqlEditor
from cutevariant.gui.querydispatcher import QueryDispatcher
from cutevariant.gui.infovariantwidget import InfoVariantWidget
from cutevariant.gui.aboutcutevariant import AboutCutevariant
from cutevariant.gui.chartquerywidget import ChartQueryWidget
from cutevariant import commons as cm
from cutevariant.commons import MAX_RECENT_PROJECTS, DIR_ICONS

# Proof of concept - testing only
from cutevariant.gui.webglquerywidget import WebGLQueryWidget
from cutevariant.gui.hpoquerywidget import HpoQueryWidget
# from cutevariant.gui.omnibar import OmniBar


LOGGER = cm.logger()


class MainWindow(QMainWindow):
    def __init__(self, parent=None):

        super(MainWindow, self).__init__()
        self.setWindowTitle("Cutevariant")
        self.toolbar = self.addToolBar("maintoolbar")
        self.toolbar.setObjectName("maintoolbar")  # For window saveState
        self.setWindowIcon(QIcon(DIR_ICONS + "app.png"))

        # Keep sqlite connection
        self.conn = None

        # Init QueryDispatcher to dispatch current query to:
        # - QueryPlugins
        # - VqlEditor
        self.query_dispatcher = QueryDispatcher()

        # Build central view based on QTabWidget
        # PS: get current view via current_tab_view()
        # Central widget encapsulates a QTabWidget and VqlEditor
        self.editor = VqlEditor()
        view_query_widget = ViewQueryWidget()

        self.tab_view = QTabWidget()
        vsplit = QSplitter(Qt.Vertical)
        vsplit.addWidget(self.tab_view)  # add QTabWidget
        vsplit.addWidget(self.editor)  # add VqlEditor
        self.setCentralWidget(vsplit)
        # Manually add query_dispatcher to VqlEditor
        self.query_dispatcher.addWidget(self.editor)
        # Add ViewQueryWidget to the QTabWidget
        self.add_tab_view(view_query_widget)
        # TODO: add other tabs here

        # Setup menubar
        self.setup_menubar()

        # Build mandatory plugins that require QueryDispatcher and menubar
        self.column_widget = ColumnQueryWidget()
        self.filter_widget = FilterQueryWidget()
        self.selection_widget = SelectionQueryWidget()
        # Add mandatory query plugins to QDockWidgets
        self.add_query_plugin(self.column_widget)
        self.add_query_plugin(self.filter_widget)
        self.add_query_plugin(self.selection_widget)
        # Testing
        # self.add_query_plugin(ChartQueryWidget())
        # self.add_query_plugin(WebGLQueryWidget())
        # self.add_query_plugin(HpoQueryWidget())

        # Setup toolbar (requires selection_widget and some actions of menubar)
        self.setup_toolbar()

        # Add mandatory variant plugin (depends on QTabWidget and menubar)
        self.info_widget = InfoVariantWidget()
        self.add_variant_plugin(self.info_widget)

        # Status Bar
        self.status_bar = QStatusBar()
        self.setStatusBar(self.status_bar)

        # Add omnibar
        # self.omnibar = OmniBar()
        # self.toolbar.addSeparator()
        # spacer = QWidget()
        # spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        # self.toolbar.addWidget(spacer)
        # self.toolbar.addWidget(self.omnibar)

        # Window geometry
        self.resize(600, 400)
        self.setGeometry(qApp.desktop().rect().adjusted(100, 100, -100, -100))

        # Load external plugins
        self.load_plugins()

        # Restores the state of this mainwindow's toolbars and dockwidgets
        self.read_settings()

        # Display messages from plugins in the status bar
        self.editor.message.connect(self.handle_plugin_message)
        view_query_widget.message.connect(self.handle_plugin_message)

        self.open("examples/test2.db")

    def add_variant_plugin(self, plugin: VariantPluginWidget):
        """Add info variant plugin to QDockWidget

        .. note:: This plugin doesn't require query_dispatcher
        .. TODO:: current tab view must send signal only for visible widget
        """
        # Connect click event on a variant in ViewQueryWidget
        # => update InfoVariantWidget
        self.current_tab_view().variant_clicked.connect(plugin.set_variant)
        self.add_panel(plugin)

    def add_query_plugin(self, plugin: QueryPluginWidget):
        """Add query plugin to QDockWidget and connect it to query_dispatcher"""
        self.query_dispatcher.addWidget(plugin)
        self.add_panel(plugin)
        plugin.message.connect(self.handle_plugin_message)

    def load_plugins(self, folder_path=None):
        """TODO ... Load plugins from path.
        What is a plugin ? A file or a module folder ?
        """
        pass

    def add_panel(self, widget, area=Qt.LeftDockWidgetArea):
        """Add given widget to a new QDockWidget and to view menu in menubar"""
        dock = QDockWidget()
        dock.setWindowTitle(widget.windowTitle())
        dock.setWidget(widget)

        # Set the objectName for a correct restoration after saveState
        dock.setObjectName(widget.objectName())
        if not widget.objectName():
            LOGGER.debug(
                "MainWindow:add_panel:: widget '%s' has no objectName attribute"
                "and will not be saved/restored",
                widget.windowTitle(),
            )
        self.addDockWidget(area, dock)
        self.view_menu.addAction(dock.toggleViewAction())

    def setup_menubar(self):
        """Menu bar setup: items and actions"""
        ## File Menu
        self.file_menu = self.menuBar().addMenu(self.tr("&File"))
        self.new_project_action = self.file_menu.addAction(
            FIcon(0xF415), self.tr("&New project"), self.new_project, QKeySequence.New
        )
        self.open_project_action = self.file_menu.addAction(
            FIcon(0xF76F),
            self.tr("&Open project ..."),
            self.open_project,
            QKeySequence.Open,
        )
        ### Recent projects
        self.recent_files_menu = self.file_menu.addMenu(self.tr("Open recent"))

        self.recentFileActions = list()
        for i in range(MAX_RECENT_PROJECTS):
            new_action = QAction()
            new_action.setVisible(False)
            # Keep actions in memory for their display to be managed later
            self.recentFileActions.append(new_action)
            self.recent_files_menu.addAction(new_action)
            new_action.triggered.connect(self.open_recent)

        # Init previous files
        self.update_recent_projects_actions()

        self.recent_files_menu.addSeparator()
        self.recent_files_menu.addAction(self.tr("Clear"), self.clear_recent_projects)

        self.file_menu.addSeparator()
        self.file_menu.addAction(
            FIcon(0xF493), self.tr("Settings ..."), self.show_settings
        )

        self.file_menu.addSeparator()
        self.file_menu.addAction(self.tr("&Quit"), qApp.quit, QKeySequence.Quit)

        ## Edit
        self.edit_menu = self.menuBar().addMenu(self.tr("&Edit"))
        self.edit_menu.addAction(FIcon(0xF18F), "&Copy", self.copy, QKeySequence.Copy)
        self.edit_menu.addAction(
            FIcon(0xF192), "&Paste", self.paste, QKeySequence.Paste
        )
        self.edit_menu.addSeparator()
        self.edit_menu.addAction(
            FIcon(0xF486), "Select all", self.select_all, QKeySequence.SelectAll
        )

        ## View
        self.view_menu = self.menuBar().addMenu(self.tr("&View"))
        self.view_menu.addAction(self.tr("Reset widgets positions"), self.reset_ui)
        self.view_menu.addSeparator()

        ## Help
        self.help_menu = self.menuBar().addMenu(self.tr("Help"))
        self.help_menu.addAction(self.tr("About Qt"), qApp.aboutQt)
        self.help_menu.addAction(self.tr("About Cutevariant"), self.aboutCutevariant)

    def setup_toolbar(self):
        """Tool bar setup: items and actions

        .. note:: Require selection_widget and some actions of Menubar
        """
        # Tool bar
        self.toolbar.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        self.toolbar.addAction(self.new_project_action)
        self.toolbar.addAction(self.open_project_action)
        self.toolbar.addSeparator()

        self.toolbar.addAction(
            FIcon(0xF412),
            self.tr("Save query"),
            self.selection_widget.save_current_query,
        )

        self.toolbar.addAction(
            FIcon(0xF40D), self.tr("Run"), self.editor.run_vql
        ).setShortcuts([Qt.CTRL + Qt.Key_R, QKeySequence.Refresh])

    def add_tab_view(self, widget):
        """Add the given widget to the current (QTabWidget),
        and connect it to the query_dispatcher"""
        self.tab_view.addTab(widget, widget.windowTitle())
        self.query_dispatcher.addWidget(widget)

    def current_tab_view(self):
        """Get the page/tab currently being displayed by the tab dialog

        :return: Return the current tab in the QTabWidget
        :rtype: <QWidget>
        """
        # Get the index position of the current tab page
        index = self.tab_view.currentIndex()
        if index == -1:
            # No tab in the widget
            return None
        return self.tab_view.currentWidget()

    def open(self, filepath):
        """Open the given db/project file

        .. note:: Called at the end of a project creation by the Wizard,
            and by Open/Open recent projects slots.

        :param filepath: Path of project file.
        :type filepath: <str>
        """
        if not os.path.exists(filepath):
            return

        # Show the project name in title and in status bar
        self.setWindowTitle("Cutevariant - %s" % os.path.basename(filepath))
        self.status_bar.showMessage(self.tr("{filepath} opened"))

        # Save directory
        app_settings = QSettings()
        app_settings.setValue("last_directory", os.path.dirname(filepath))

        # Create connection
        self.conn = get_sql_connexion(filepath)

        # Create a query
        query = Query(self.conn)

        # Dispatch the query to all widget from the router
        self.query_dispatcher.query = query

        # Update all widgets
        self.query_dispatcher.update_all_widgets()

        # Refresh recent opened projects
        self.adjust_recent_projects(filepath)

    def adjust_recent_projects(self, filepath):
        """Adjust the number of of recent projects to display

        .. note:: Called after a successful file opening.

        :param filepath: Path of project file.
        :type filepath: <str>
        """
        # Get recent projects list
        recent_projects = self.get_recent_projects()

        try:
            recent_projects.remove(filepath)
        except ValueError:
            # New file is not already in the list
            pass
        # Prepend new file
        recent_projects = [filepath] + recent_projects
        recent_projects = (
            recent_projects
            if len(recent_projects) <= MAX_RECENT_PROJECTS
            else recent_projects[:-1]
        )

        # Save in settings
        app_settings = QSettings()
        app_settings.setValue("recent_projects", recent_projects)

        # Display
        self.update_recent_projects_actions()

    def update_recent_projects_actions(self):
        """Display recent projects in the menu

        .. note:: Called after a successful file opening and during the launch
            of the software.
        """
        # Get recent projects list
        recent_projects = self.get_recent_projects()

        index = -1
        for index, filepath in enumerate(recent_projects, 0):
            action = self.recentFileActions[index]
            # Get filename
            # TODO: get project name ?
            action.setText(os.path.basename(filepath))
            action.setData(filepath)
            action.setVisible(True)

        # Display the action
        if index == -1:
            self.recent_files_menu.setEnabled(False)
        else:
            self.recent_files_menu.setEnabled(True)

        # Switch off useless actions
        # index = -1 if there is no recent_projects
        index = 0 if index < 0 else index + 1
        for i in range(index, MAX_RECENT_PROJECTS):
            self.recentFileActions[i].setVisible(False)

    def get_recent_projects(self):
        """Return the list of recent projects stored in settings

        :return: List of recent projects.
        :type: <list>
        """
        # Reload last projects opened
        app_settings = QSettings()
        recent_projects = app_settings.value("recent_projects", list())

        # Check if recent_projects is a list() (as expected)
        if isinstance(recent_projects, str):
            recent_projects = [recent_projects]

        return recent_projects

    def handle_plugin_message(self, message):
        """Slot to display message from plugin in the status bar"""
        self.status_bar.showMessage(message)

    def new_project(self):
        """Slot to allow creation of a project with the Wizard"""
        wizard = ProjectWizard()
        if wizard.exec_():
            db_filename = (
                wizard.field("project_path")
                + QDir.separator()
                + wizard.field("project_name")
                + ".db"
            )
            try:
                self.open(db_filename)
            except Exception as e:
                self.status_bar.showMessage(e.__class__.__name__ + ": " + str(e))
                raise

    def open_project(self):
        """Slot to open an already existing project"""
        # Reload last directory used
        app_settings = QSettings()
        last_directory = app_settings.value("last_directory", QDir.homePath())

        filepath, _ = QFileDialog.getOpenFileName(
            self,
            self.tr("Open project"),
            last_directory,
            self.tr("Cutevariant project (*.db)"),
        )
        if filepath:
            try:
                self.open(filepath)
            except Exception as e:
                self.status_bar.showMessage(e.__class__.__name__ + ": " + str(e))
                raise

    def open_recent(self):
        """Slot to load a recent project"""
        action = self.sender()
        self.open(action.data())

    def clear_recent_projects(self):
        """Slot to clear the list of recent projects"""
        app_settings = QSettings()
        app_settings.remove("recent_projects")
        self.update_recent_projects_actions()

    def show_settings(self):
        """Slot to show settings window"""
        widget = SettingsWidget()
        widget.exec()

    def aboutCutevariant(self):
        """Slot to show about window"""
        dialog_window = AboutCutevariant()
        dialog_window.exec()

    def reset_ui(self):
        """Slot to reset the position of docks to the state of the previous launch"""
        # Set reset ui boolean (used by closeEvent)
        self.requested_reset_ui = True

        # Restore docks
        app_settings = QSettings()
        self.restoreState(QByteArray(app_settings.value("windowState")))

    def copy(self):
        pass

    def paste(self):
        pass

    def select_all(self):
        """Select all elements in the current tab's view"""
        self.current_tab_view().view.selectAll()

    def closeEvent(self, event):
        """Save the current state of this mainwindow's toolbars and dockwidgets

        .. warning:: Make sure that the property objectName is unique for each
            QToolBar and QDockWidget added to the QMainWindow.

        .. note:: Reset windowState if asked.
        """
        app_settings = QSettings()

        if self.requested_reset_ui:
            # Delete window state setting
            app_settings.remove("windowState")
        else:
            app_settings.setValue("geometry", self.saveGeometry())
            #  TODO: handle UI changes by passing UI_VERSION to saveState()
            app_settings.setValue("windowState", self.saveState())

        super().closeEvent(event)

    def read_settings(self):
        """Restore the state of this mainwindow's toolbars and dockwidgets

        .. note:: If windowState is not stored, current state is written.
        """
        # Init reset ui boolean
        self.requested_reset_ui = False

        app_settings = QSettings()
        self.restoreGeometry(QByteArray(app_settings.value("geometry")))
        #  TODO: handle UI changes by passing UI_VERSION to saveState()
        window_state = app_settings.value("windowState")
        if window_state:
            self.restoreState(QByteArray(window_state))
        else:
            # Setting has been deleted: set the current default state
            #  TODO: handle UI changes by passing UI_VERSION to saveState()
            app_settings.setValue("windowState", self.saveState())



from abc import ABC, abstractclassmethod


class AbstractReader(ABC):
    """Base class for all Readers required to import variants into the database.
    Subclass it if you want a new file parser .

    Attributes:
        device: a file object typically returned by open()
        file_size: file size in bytes
        read_bytes: current bytes readed (progression = read_bytes / file_size)
        samples: list of samples in the file (default: empty)

    Example:
        with open(filename,"r") as file:
            reader = Reader()
            reader.get_variants()
    """

    def __init__(self, device):
        super(AbstractReader, self).__init__()
        self.device = device
        self.file_size = 0
        self.read_bytes = 0
        self.samples = []
        self.fields = tuple()

    @abstractclassmethod
    def get_variants(self):
        """Abstract method that must return variants as an iterable of dictionnaries.

        Mandatory fields:
        =================
        `chr`, `pos`, `ref`, `alt` are mandatory.
        :Example:
            [
                {"chr": "chr3","pos": 3244,"ref": "A","alt":"C"},
                {"chr": "chr4","pos": 3244,"ref": "A","alt":"C"},
                {"chr": "chr5","pos": 3244,"ref": "A","alt":"C"},
            ]

        Full fields:
        ============
       `annotations`, `samples` and other fields can be added.
        :Example:
            [{
                "chr": "chr3",
                "pos": 3244,
                "ref": "A",
                "alt":"C",
                "field_n": "value_n",
                "annotations": [
                    {"gene": "GJB2", "transcripts": "NM_00232.1", "field_n": "value_n"},
                    {"gene": "GJB2", "transcripts": "NM_00232.2", "field_n": "value_n"}
                ],
                "samples": [
                    {"name":"boby", "gt": 1, "field_n":"value_n"},
                    {"name":"kevin", "gt": 1, "field_n":"value_n"}
                ]
            },]

        :return: A generator of variants
        :rtype: <generator>
        """
        raise NotImplementedError(self.__class__.__name__)

    @abstractclassmethod
    def get_fields(self):
        """Abstract methodthat must return fields description defined from parse_variant output.

        You **must** define sqlite type for each field (text, integer, bool).

        This function should set the attribute `fields` because it is called
        several times during the import and it could be expensive to redo the
        parsing every time.

        This function is called a first time before variants insertion.

        Full output:
        ============
        [
        {"name": "chr", "type": "text", "category": "variant", "description": "description"},
        {"name": "pos", "type": "text", "category": "variant", "description": "description"},
        {"name": "ref", "type": "text", "category": "variant", "description": "description"},
        {"name": "alt", "type": "text", "category": "variant", "description": "description"},
        {"name": "field_n", "type": "text", "category": "variant", "description": "description"},
        {"name": "name", "type": "text", "category": "annotations", "samples": "description"},
        {"name": "gt", "type": "text", "category": "annotations", "samples": "description"}
        ]

        :return: A generator of fields
        :rtype: <generator>
       """
        raise NotImplementedError(self.__class__.__name__)

    def get_fields_by_category(self, category: str):
        """Syntaxic suggar to get fields according their category

        :param category can be usually variants, samples, annotations
        :return: A generator of fields
        :rtype: <generator>
        """
        return (field for field in self.get_fields() if field["category"] == category)

    def get_variants_count(self) -> int:
        """Get variant count from the device.
        Override this method to make it faster
        """
        return len(tuple(self.get_variants()))

    def get_samples(self) -> str:
        """Return list of samples.
        Override this method to have samples in sqlite database.
        """
        return self.samples

    def get_extra_fields(self):
        """Mandatory fields to add automatically

        ..todo: Move this methods somewhere else ..
        ..warning: DEPRECTATED
        """
        yield from self.parse_fields()
        yield {
            "name": "description",
            "type": "text",
            "category": "extra",
            "description": "description of variant",
        }
        yield {
            "name": "favoris",
            "type": "bool",
            "category": "extra",
            "description": "is favoris",
            "default": False,
        }


def check_variant_schema(variant: dict):
    """Test if get_variant returns well formated nested data.

    This method is for testing purpose. It raises an exception if data is corrupted

    :param variant dict returned by AbstractReader.get_variant()

    """
    try:
        from schema import Schema, And, Or, Use, Optional, Regex
    except ImportError as e:
        print("You should install optional package 'schema' via:")
        print("\t - pypi: pip install cutevariant[dev]")
        print("\t - git repo in editable mode: pip -e . [dev]")
        raise e

    checker = Schema(
        {
            "chr": And(Use(str.lower), str),
            "pos": int,
            "ref": And(Use(str.upper), Regex(r"^[ACGTN]+")),
            "alt": And(Use(str.upper), Regex(r"^[ACGTN]+")),
            Optional(str): Or(int, str, bool, float, None),
            Optional("annotations"): [
                {
                    "gene": str,
                    "transcript": str,
                    Optional(str): Or(int, str, bool, float),
                }
            ],
            Optional("samples"): [
                {"name": str, "gt": And(int, lambda x: x in [-1, 0, 1, 2])}
            ],
        }
    )

    checker.validate(variant)


def check_field_schema(field: dict):
    """Test if get_field returns well formated data

    This method is for testing purpose. It raises an exception if data is corrupted

    :param field dict returned by AbstractReader.get_field()
    """
    try:
        from schema import Schema, And, Use, Optional
    except ImportError as e:
        print("You should install optional package 'schema' via:")
        print("\t - pypi: pip install cutevariant[dev]")
        print("\t - git repo in editable mode: pip -e . [dev]")
        raise e

    checker = Schema(
        {
            "name": And(str, Use(str.lower)),
            "type": lambda x: x in ["str", "int", "bool", "float"],
            "category": lambda x: x in ["variants", "annotations", "samples"],
            "description": str,
            Optional("constraint", default="NULL"): str,
        }
    )

    checker.validate(field)



# Standard imports
import re
import sqlite3
from functools import lru_cache

# Custom imports
from . import sql
from . import vql
from cutevariant.commons import logger, DEFAULT_SELECTION_NAME

LOGGER = logger()

_GENOTYPE_FUNCTION_NAME = "genotype"
_PHENOTYPE_FUNCTION_NAME = "phenotype"


class Query:
    """Class used to build sql query to select variants records

    Available attributes:
        conn (sqlite3.Connection)
        columns (list of str and tuple): Fields names from variants and annotations table (Select clause)
        filter (dict): Hierarchical dictionnary to filter variants (Where clause)
        selection (str): Virtual table of variants (From clause)
        order_by(str): Order variants by a specific column
        group_by(list of str): Group variants by columns

    About functions:
        `columns` and `filter` can contains function as tuple.
        A function is defined by:
            - function name
            - arguments (sample name, etc.)
            - sql field name

        For example :

            query.columns = ["chr","pos", ("genotype","boby","GT")]

        is equivalent to:

            SELECT chr, pos, genotype("boby").gt

    Example:
        conn = sqlite.connection(":memory")
        query = Query(conn)
        query.columns = ["chr","pos","ref","alt"]
        query.selection = "variants"
        query.filter = {"AND": [
                {"field": "ref", "operator": "=", "value": "A"},
                {
                    "OR": [
                        {"field": "chr", "operator": "=", "value": "chr5"},
                        {"field": "chr", "operator": "=", "value": "chr3"},
                    ]
                },}}

        conn.execute(conn.sql())
    """

    def __init__(
        self,
        conn,
        columns=["chr", "pos", "ref", "alt"],
        filter=dict(),
        selection=DEFAULT_SELECTION_NAME,
        group_by=["chr", "pos", "ref", "alt"],
    ):
        self.conn = conn
        self.columns = columns
        self.filter = filter
        self.selection = selection
        self.order_by = None
        self.order_desc = True
        self.group_by = group_by

        self._samples_to_join = set()

        # Mapping cols => table
        # Get columns description from the given table
        tables = ("variants", "annotations")
        self.col_table_mapping = {
            table_name: set(sql.get_columns(self.conn, table_name))
            for table_name in tables
        }

    def extract_samples_from_columns_and_filter(self, filter_only=False):
        """Extract samples if columns or filter contains function.

        The aim is to dynamically add JOIN clauses on 'samples' table.
        `self._samples_to_join` is modified here.

        .. note:: About functions:
            `columns` and `filter` can contains function as tuple.
            A function is defined by:
                - function name
                - arguments (sample name, etc.)
                - sql field name

            This columns selection can be writted in VQL as follow:

                SELECT chr, pos, genotype("boby").gt

            "boby" will be added to `self._samples_to_join`

        .. note:: `_samples_to_join` is a dict with sample_names as keys
            and sample_ids as values.
        """
        # Set sample names to join by searching them in columns
        if not filter_only:
            columns_in_columns = self.get_samples_names_from_functions(
                self.columns, _GENOTYPE_FUNCTION_NAME
            )
        else:
            columns_in_columns = set()

        # Set sample names to join by searching them in filter
        # iter_filter(): Recursive loop over filter to extract field name only
        columns_in_filter = self.get_samples_names_from_functions(
            self.iter_filter(self.filter), _GENOTYPE_FUNCTION_NAME
        )

        # Get samples required by columns and filter
        required_samples_names = columns_in_columns | columns_in_filter

        self._samples_to_join = {
            sample["name"]: sample["id"]
            for sample in sql.get_samples(self.conn)
            if sample["name"] in required_samples_names
        }

        LOGGER.debug("DETECT %s in %s", self._samples_to_join.keys(), self.columns)

    def get_samples_names_from_functions(self, columns, function_name):
        """Get samples names from given functions if their function_name matches
        to the given one.

        :param columns: Iterable of columns in which functions are searched.
        :param function_name: Function name to be search in functions.
        :return: Set of arguments (samples names).
        :rtype: <set>
        """
        samples = set()
        # Get function tuples
        #  A function is a tuple with 3 elements.
        #  The second element is the sample name
        # TODO: is test on length usefull?
        functions = (col for col in columns if isinstance(col, tuple) and len(col) == 3)

        for function in functions:
            function_name, function_argument, field_name = function

            if function_name == function_name:
                # function_argument is a sample_name here
                samples.add(function_argument)
        return samples

    def detect_annotations_table_requirement(self, filter_only=False):
        """Return True if annotations table is required after searching in
        columns and / or filter.

        :key filter_only: If True, the search will be made only in 'self.filter'
        :type filter_only: <boolean>
        :rtype: <boolean>
        """
        # Get columns in filter
        cols_in_annotations = {
            col for col in self.iter_filter(self.filter)
        } & self.col_table_mapping["annotations"]
        if cols_in_annotations:
            LOGGER.debug(
                "detect_annotations_table_requirement: found col in filter: %s",
                cols_in_annotations,
            )
            return True

        if filter_only:
            # Stop here, no col has been found in filter => False
            return False

        # Get columns in columns
        cols_in_annotations = set(self.columns) & self.col_table_mapping["annotations"]
        if cols_in_annotations:
            LOGGER.debug(
                "detect_annotations_table_requirement: found col in columns: %s",
                cols_in_annotations,
            )
            return True
        return False

    def iter_filter(self, node):
        """Recursive loop over filter to extract field name only
        Recall: {'AND': [{'field': 'ref', 'operator': 'IN', 'value': "('A', 'T', 'G', 'C')"},
        {'field': 'alt', 'operator': 'IN', 'value': "('A', 'T', 'G', 'C')"}]}
        Aim: yield columns involved in filter
        """
        if isinstance(node, dict) and len(node) == 3:
            yield node["field"]

        if isinstance(node, dict):
            for i in node:
                yield from self.iter_filter(node[i])

        if isinstance(node, list):
            for i in node:
                yield from self.iter_filter(i)

    ##--------------------------------------------------------------------------

    def get_columns(self, do_not_add_default_things=False):
        """Get list of columns ready to be inserted in a query string
        .. seealso:: sql() or sql_count()
        :rtype: <list>
       """
        # Set default columns if columns is empty
        # Otherwise, columns are kept unmodified
        if not self.columns:
            self.columns = ["chr", "pos", "ref", "alt"]

        if do_not_add_default_things:
            # Keep columns as they are set in the query
            sql_columns = []
        else:
            # Some queries must have this field in addition to theirs
            sql_columns = ["variants.id"]

        # Replace genotype function by name
        # Transform ("genotype", "boby","gt") to "`gt_boby`.gt" to perform SQL JOIN
        for col in self.columns:
            if isinstance(col, tuple):
                function_name, arg, field_name = col
                if function_name == _GENOTYPE_FUNCTION_NAME:
                    # Secure column name
                    col = f"`gt_{arg}`.{field_name}"

            sql_columns.append(col)

        # If 'group by', add extra columns (child count and child ids)
        # Required for viewquerywidget.py
        if not do_not_add_default_things and self.group_by:
            sql_columns.extend(["COUNT(*) as 'children'"])

        return sql_columns

    def get_joints(self, do_not_add_default_things=False, filter_only=False):
        """Get string of joints ready to be appended to a query string
        .. seealso:: sql() or sql_count()
        :key filter_only: If True, the search will be made only in 'self.filter'
        :type filter_only: <boolean>
        :rtype: <str>
        """
        # On filter and columns
        # Joint on annotations is mandatory for display queries
        # We do not do joint if it is explicitly forbidden (in this case
        # a join is automatically decided if it is required by filter or columns)
        is_col_in_annotations = self.detect_annotations_table_requirement(
            filter_only=filter_only
        )
        query = (
            ""
            if do_not_add_default_things and not is_col_in_annotations
            else " LEFT JOIN annotations ON annotations.variant_id = variants.id"
        )

        if self.selection and self.selection != DEFAULT_SELECTION_NAME:
            # Add jointure with 'selections' table
            query += f"""
            INNER JOIN selection_has_variant sv ON sv.variant_id = variants.id
            INNER JOIN selections s ON s.id = sv.selection_id AND s.name = '{self.selection}'
            """

        #  Add Join on sample_has_variant
        #  This is done if genotype() function has been found in columns or filter,
        # or filter only if specified.
        # _samples_to_join is a dict with sample_names as keys and sample_ids as values
        self.extract_samples_from_columns_and_filter(filter_only=filter_only)
        for sample_name, sample_id in self._samples_to_join.items():
            query += f"""
            LEFT JOIN sample_has_variant `gt_{sample_name}`
             ON `gt_{sample_name}`.variant_id = variants.id
             AND `gt_{sample_name}`.sample_id = {sample_id}
            """
        return query

    def sql(self, limit=0, offset=0, do_not_add_default_things=False) -> str:
        """Build a sql query according to attributes for raw and display queries

        .. note:: Some queries require that this functions doesn't automatically
            add default columns or joins.
            In this case the argument 'do_not_add_default_things' must be set to True.

            - `variants.id` will not be appended to SELECT clause.
            - A LEFT JOIN on annotations will not be made if it is not required
            by a filter or by columns.
            Typical query concerned: see ChartQueryWidget:on_change_query()
            In other cases, 'annotations' table is mandatory for display queries
            which group variants on transcripts for example.

        .. note:: This function can be called as this, but its main purpose is
            to be used by self.items().

        .. seealso:: sql_count()

        :param limit: SQL LIMIT for pagination
        :param offset: SQL OFFSET for pagination
        :key do_not_add_default_things: See previous note (default: False)
        :return: A SQL query ready to be executed.
        :rtype: <str>
        """
        ## Build columns
        sql_columns = self.get_columns(do_not_add_default_things)
        query = f"SELECT {','.join(sql_columns)} "

        ## Add FROM clause
        # Explicitly query all variants + ...
        # do_not_add_default_things set to True to avoid groups of variants
        # when there is no checked column in annotations.
        query += "FROM variants" + self.get_joints(do_not_add_default_things=True)

        ## Add WHERE filter
        if self.filter:
            query += " WHERE " + self.filter_to_sql(self.filter)

        ## Add GROUP BY command
        if self.group_by:
            query += " GROUP BY " + ",".join(self.group_by)

        ## Add ORDER BY command
        if self.order_by is not None:
            direction = "DESC" if self.order_desc else "ASC"
            query += f" ORDER BY {self.order_by} {direction}"

        ## Add LIMIT and OFFSET clauses
        if limit > 0:
            query += f" LIMIT {limit} OFFSET {offset}"

        LOGGER.debug("Query:sql:: query: %s", query)
        return query

    def sql_count(self):
        """Build a query for count aggregation queries

        Aim: Build queries with chunks of joints/group by commands that
        are only required based on the columns and filters used.

        Useful chunks:
            - Joints:
                - annotations: Only if a column in filter is stored in that table.
                We don't care if such a column is used in SELECT clause.
                It's a time saving and a generator of cleaner queries.
                => decided by detect_annotations_table_requirement(filter_only=True)
                - selections: Only if a selection is selected, and if this is not
                the default one.
                - samples: Only if a sample function is used in filter clause.
                We don't care if such a column is used in SELECT clause.
                It's a time saving and a generator of cleaner queries.
            - Group by:
                - (chr, pos, ref, alt) is the primary key so there will be as many
                groups as variants in the database.
                => no group by is required in this case.
                - (chr, pos) is a group by that modifies the number of variants
                because it modifies the definition of the unicity of a variant.
                => group by is required to reduce the number of variants.

        .. seealso:: sql()
        """
        ## Build columns
        query = "SELECT variants.id "

        ## Add FROM clause
        # Search annotations need on filter only
        # Explicitly query all variants + ...
        # do_not_add_default_things is True to force the removing of annotations
        # as much as possible
        query += "FROM variants" + self.get_joints(
            do_not_add_default_things=True, filter_only=True
        )

        ## Add WHERE filter
        if self.filter:
            query += " WHERE " + self.filter_to_sql(self.filter)

        ## Add GROUP BY command
        if self.group_by != ["chr", "pos", "ref", "alt"]:
            # (chr, pos) is a group by that modifies the number of variants
            # because it modifies the definition of the unicity of a variant.
            # (chr, pos, ref, alt) is the primary key so there will be as many
            # groups as variants in the database.
            query += " GROUP BY " + ",".join(self.group_by)

        LOGGER.debug("Query:sql_count:: query: %s", query)
        return query

    ##--------------------------------------------------------------------------

    def items(self, limit=0, offset=0):
        """Execute SQL query and return variants as a list

        .. note:: Used:
            - as dict of variants in chartquerywidget.py
            - as tuples of variants in viewquerywidget.py

        :param limit: SQL LIMIT for pagination
        :param offset: SQL OFFSET for pagination
        :return: Generator of variants as sqlite3.Row objects.
            A Row instance serves as a highly optimized row_factory for
            Connection objects. It tries to mimic a tuple in most of its features.

            It supports mapping access by column name and index, iteration,
            representation, equality testing and len().
        :rtype: <generator <sqlite3.Row>>

        :Example:

            for row in query.items():
        ...     print(tuple(row))
        (324, "chr2", "24234", "A", "T", ...)
        ...     print(dict(row))
        {"rowid":23423, "chr":"chr2", "pos":4234, "ref":"A", "alt": "T", ...}

        """
        self.conn.row_factory = sqlite3.Row
        yield from self.conn.execute(self.sql(limit, offset))

    def filter_to_sql(self, node: dict) -> str:
        """Recursive function to convert the self.filter hierarchical dictionnary
        into a SQL WHERE clause.

        :param node: hierachical dictionnary
        :return: a SQL WHERE clause

        ..seealso: filter

        :Example of filter:
        filter: {'AND': [
            {'field': 'ref', 'operator': 'IN', 'value': "('A', 'T', 'G', 'C')"},
            {'field': 'alt', 'operator': 'IN', 'value': "('A', 'T', 'G', 'C')"}
        ]}
        """

        if not node:
            return ""

        # Function to detect IF node is a Condition node (AND/OR)
        # OR a field node with (name, operator, value) as keys
        is_field = lambda x: True if len(x) == 3 else False

        if is_field(node):
            # print("IS FIELD", node)

            # Process value
            value = node["value"]
            operator = node["operator"]
            field = node["field"]

            # TODO ... c'est degeulasse ....

            if operator in ("IN", "NOT IN"):
                # DO NOT enclose value in quotes
                # node: {'field': 'ref', 'operator': 'IN', 'value': "('A', 'T', 'G', 'C')"}
                # wanted: ref IN ('A', 'T', 'G', 'C')
                pass

            elif isinstance(value, list):
                value = "(" + ",".join(value) + ")"

            else:
                value = str(value)

            # Process field
            if isinstance(field, tuple) and len(field) == 3:
                #  Function ? ("genotype","sample","gt")
                fct, arg, f = field
                field = f"`gt_{arg}`.{f}"

            # There must be spaces between these strings because of strings operators (IN, etc.)
            return "%s %s %s" % (field, operator, value)

        else:
            # Not a field: 1 key only: the logical operator
            logic_op = list(node.keys())[0]
            # Recursive call for each field in the list associated to the
            # logical operator.
            # node:
            # {'AND': [
            #   {'field': 'ref', 'operator': 'IN', 'value': "('A', 'T', 'G', 'C')"},
            #   {'field': 'alt', 'operator': 'IN', 'value': "('A', 'T', 'G', 'C')"}
            # ]}
            # Wanted: ref IN ('A', 'T', 'G', 'C') AND alt IN ('A', 'T', 'G', 'C')
            out = [self.filter_to_sql(child) for child in node[logic_op]]
            # print("OUT", out, "LOGIC", logic_op)
            # OUT ["refIN'('A', 'T', 'G', 'C')'", "altIN'('A', 'T', 'G', 'C')'"]

            return "(" + f" {logic_op} ".join(out) + ")"

    ##--------------------------------------------------------------------------

    def create_selection(self, name):
        """Store variant set from the current query into `selections` table

        :param name: Name of the selection

        :Example:

            query1 = Query(conn)
            query1.from_vql("SELECT chr, pos FROM variants WHERE chr = 3 ")
            print(query1.variants_count())
        10
            query1.create_selection("boby")
            query2 = Query(conn)
            query2.from_vql("SELECT chr, pos FROM boby")
            print(query2.variants_count())
        10 # same as before

        :return: The id of the new selection in the database. None in case of error.
        :rtype: <int> or None
        """
        # TODO: handle by key argument
        return sql.create_selection_from_sql(self.conn, self.sql(), name=name)

    ##--------------------------------------------------------------------------

    @lru_cache(maxsize=128)
    def _cached_variants_count_query(self, sql_query):
        """Wrapped function with a memoizing callable that saves up to the
        maxsize most recent calls.

        .. note:: The LRU feature performs best when maxsize is a power-of-two.

        .. note:: The COUNT() aggregation function is expensive on partially
            indexed tables (because dynamically built) for large dataset
            and it seems difficult to predict which fields will be requested
            by the user.
        """
        # Trick to accelerate UI refresh on basic queries
        if (
            (not self.selection or self.selection == DEFAULT_SELECTION_NAME)
            and not self.filter
            and self.group_by == ["chr", "pos", "ref", "alt"]
        ):
            LOGGER.debug("SELECT MAX(...")
            return self.conn.execute(
                "SELECT MAX(variants.id) as count FROM variants"
            ).fetchone()[0]

        LOGGER.debug(sql_query)
        return self.conn.execute(
            f"SELECT COUNT(*) as count FROM ({sql_query})"
        ).fetchone()[0]

    def variants_count(self) -> int:
        """Return variant count from the current query

        .. note:: This function is used in:
            - viewquerywidget.py in `load()` to keep the total of variants for
            paging purposes of the interface.
        """
        LOGGER.debug("Query:variants_count:: query:")
        count = self._cached_variants_count_query(self.sql_count())
        LOGGER.debug(
            "Query:variants_count:: %s", self._cached_variants_count_query.cache_info()
        )
        return count

    ##--------------------------------------------------------------------------

    def __repr__(self):
        return f"""
        columns : {self.columns}
        filter: {self.filter}
        selection: {self.selection}
        """

    ##--------------------------------------------------------------------------
    def to_vql(self) -> str:
        """Build a VQL query from the current Query

        .. seealso:: from_vql()
        .. todo:: Should be a static method

        :return: A VQL query
        :rtype: <str>
        """

        # TODO : move all VQL to VQLEDItor
        #  DEGEU.. pour tester juste
        _c = []
        for col in self.columns:
            if isinstance(col, tuple):
                fct, arg, field = col
                if fct == _GENOTYPE_FUNCTION_NAME:
                    col = f'genotype("{arg}")'
            _c.append(col)

        base = f"SELECT {','.join(_c)} FROM {self.selection}"
        where = ""
        if self.filter:
            where = f" WHERE {self.filter_to_sql(self.filter)}"
        return base + where

    ##--------------------------------------------------------------------------

    def check(self):
        """Return True if query is valid"""
        raise NotImplementedError



# Qt imports
from PySide2.QtCore import QObject

# Custom imports
from .plugin import QueryPluginWidget
from cutevariant.core import Query
from cutevariant.commons import logger

LOGGER = logger()


class QueryDispatcher(QObject):
    """Redirect query between widgets

    If one of the widget emits a `changed` signal, then all belonging widgets
    change their query except the sender.
    """

    def __init__(self):
        super().__init__()
        self.widgets = []
        self._query = None

    def addWidget(self, widget: QueryPluginWidget):
        """Add a widget into the router

        :param widget: a query widget and connect its `query_changed` event to
            the slot `update_all_widgets()`.
        """
        self.widgets.append(widget)
        widget.query_changed.connect(self.update_all_widgets)

    @property
    def query(self):
        """Get current query
        :return: Return current query
        """
        return self._query

    @query.setter
    def query(self, query: Query):
        """Set query and update the query of all widgets

        :param query: a Query object
        """
        self._query = query
        for widget in self.widgets:
            widget.query = query

    def update_all_widgets(self):
        """Dispatch a `query_changed` triggered from one widget to all widgets
        except the sender and the invisible widgets.
        """
        # Call on_query_changed of all referenced widgets
        for widget in self.widgets:
            if widget != self.sender() and widget.isVisible():
                LOGGER.debug(
                    "QueryDispatcher:update_all_widgets:: change event for %s",
                    widget.__class__.__name__,
                )
                widget.on_change_query()



"""A place to store style rules for the GUI"""
from cutevariant.commons import DIR_STYLES
from PySide2.QtGui import QPalette, QColor

TYPE_COLORS = {
    "str": "#27A4DD",  # blue
    "bool": "#F1646C",  # red
    "float": "#9DD5C0",  # light blue
    "int": "#FAC174",  # light yellow
    "NoneType": "#FFFFFF",  # white
}

IMPACT_COLOR = {
    "LOW": "#71E096",
    "MODERATE": "#F5A26F",
    "HIGH": "#ed6d79",
    "MODIFIER": "#55abe1",
}

GENE_COLOR = "#F5A26F"
WARNING_BACKGROUND_COLOR = "#FFCCBA"
WARNING_TEXT_COLOR = "#D73705"


# Sequence ontology colors
SO_COLOR = {
    "stop": "#808080",
    "utr": "#911EB4",
    "splice": "#F58231",
    "intron": "#0082C8",
    "intergenic": "#3CB44B",
    "frameshift": "#E6194B",
    "missense": "#F032E6",
}


def dark(app):
    """ Apply Dark Theme to the Qt application instance.
        Args:
            app (QApplication): QApplication instance.
    """



    darkPalette = QPalette()

    # base
    darkPalette.setColor(QPalette.WindowText, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.Button, QColor(53, 53, 53))
    darkPalette.setColor(QPalette.Light, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.Midlight, QColor(90, 90, 90))
    darkPalette.setColor(QPalette.Dark, QColor(35, 35, 35))
    darkPalette.setColor(QPalette.Text, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.BrightText, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.ButtonText, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.Base, QColor(42, 42, 42))
    darkPalette.setColor(QPalette.Window, QColor(53, 53, 53))
    darkPalette.setColor(QPalette.Shadow, QColor(20, 20, 20))
    darkPalette.setColor(QPalette.Highlight, QColor(42, 130, 218))
    darkPalette.setColor(QPalette.HighlightedText, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.Link, QColor(56, 252, 196))
    darkPalette.setColor(QPalette.AlternateBase, QColor(66, 66, 66))
    darkPalette.setColor(QPalette.ToolTipBase, QColor(53, 53, 53))
    darkPalette.setColor(QPalette.ToolTipText, QColor(180, 180, 180))

    # disabled
    darkPalette.setColor(QPalette.Disabled, QPalette.WindowText,
                         QColor(127, 127, 127))
    darkPalette.setColor(QPalette.Disabled, QPalette.Text,
                         QColor(127, 127, 127))
    darkPalette.setColor(QPalette.Disabled, QPalette.ButtonText,
                         QColor(127, 127, 127))
    darkPalette.setColor(QPalette.Disabled, QPalette.Highlight,
                         QColor(80, 80, 80))
    darkPalette.setColor(QPalette.Disabled, QPalette.HighlightedText,
                         QColor(127, 127, 127))

    app.setPalette(darkPalette)

    #_apply_base_theme(app)
    with open(DIR_STYLES + "dark.style.qss","r") as file:
        app.setStyleSheet(file.read())


# Qt imports
from PySide2.QtCore import qApp, Qt, QRect, QPoint, QBuffer, QByteArray
from PySide2.QtGui import (
    QIconEngine,
    QColor,
    QPainter,
    QIcon,
    QFontDatabase,
    QFont,
    QPalette,
    QPixmap,
    QPen,
)

# Custom imports
from cutevariant.commons import logger

LOGGER = logger()


class FIconEngine(QIconEngine):
    """Base class for FIcon; used to load custom font"""

    # Class attribute
    font = None

    def __init__(self):
        super().__init__()
        self.setColor(qApp.palette().color(QPalette.Normal, QPalette.Text))

    def setCharacter(self, hex_character: int):
        self.hex_character = hex_character

    def setColor(self, color: QColor):
        self.color = color

    def paint(
        self, painter: QPainter, rect: QRect, mode: QIcon.Mode, state: QIcon.State
    ):
        """override"""
        font = FIconEngine.font if hasattr(FIconEngine, "font") else painter.font()
        painter.save()

        if mode == QIcon.Disabled:
            painter.setPen(QPen(qApp.palette().color(QPalette.Disabled, QPalette.Text)))

        else:
            painter.setPen(QPen(self.color))

        font.setPixelSize(rect.size().width())

        painter.setFont(font)
        # painter.setRenderHint(QPainter.HighQualityAntialiasing, True)
        painter.drawText(
            rect, Qt.AlignCenter | Qt.AlignVCenter, str(chr(self.hex_character))
        )
        painter.restore()

    def pixmap(self, size, mode, state):
        pix = QPixmap(size)
        pix.fill(Qt.transparent)
        painter = QPainter(pix)
        self.paint(painter, QRect(QPoint(0, 0), size), mode, state)
        return pix

    @classmethod
    def setFontPath(cls, filename):
        """Set font from a font file and return it

        :return: QFont or False if the file is not loaded
        """
        db = QFontDatabase()
        font_id = db.addApplicationFont(filename)
        if font_id == -1:
            LOGGER.error("FIconEngine:setFontPath:: cannot load font icon.")
            return False

        cls.font = QFont(db.applicationFontFamilies(font_id)[0])
        return cls.font


class FIcon(QIcon):
    """Handy public class to load and use custom font in QIcons"""

    def __init__(self, hex_character: int, color=None):
        """Build an icon with the given character and color from the current font"""
        self.engine = FIconEngine()
        self.engine.setCharacter(hex_character)
        if color:
            self.engine.setColor(color)
        else:
            self.engine.setColor(qApp.palette().text().color())
        super().__init__(self.engine)

    def to_base64(self, w=32, h=32):
        """Return icon as base64 to make it work with html"""
        pix = self.pixmap(w, h)
        data = QByteArray()
        buff = QBuffer(data)
        pix.save(buff, "PNG")
        return data.toBase64().data().decode()


def setFontPath(filename):
    """Handy function to load font file

    .. note:: Fonts are supposed to be in cm.DIR_FONTS
    .. note:: This function is called only 1 time at the start of the program
    """
    return FIconEngine.setFontPath(filename)



# Standard imports
import os
import csv
import sqlite3

# Custom imports
from .reader.abstractreader import AbstractReader
from .readerfactory import create_reader
from .sql import *


def async_import_reader(conn, reader: AbstractReader, **kwargs):
    """Import data via the given reader into a SQLite database via the given connection

    :param conn: sqlite connection
    :param reader: must be a AbstractReader base class
    :return: yield progression and message
    :rtype: <generator <int>, <str>>
    """
    # Create project
    yield 0, f"Importing data with {reader}"
    create_project(
        conn,
        name=kwargs.get("project_name", "UKN"),
        reference=kwargs.get("reference", "UKN"),
    )

    yield 0, "Creating table shema..."
    # Create table fields
    create_table_fields(conn)

    # # Create table samples
    create_table_samples(conn)

    # Create annotations tables
    create_table_annotations(conn, reader.get_fields_by_category("annotations"))

    # Create variants tables
    create_table_variants(conn, reader.get_fields_by_category("variants"))

    # Create selection
    create_table_selections(conn)

    # Insert samples
    yield 0, "Inserting samples..."
    insert_many_samples(conn, reader.get_samples())

    # Insert fields
    yield 0, "Inserting fields..."
    insert_many_fields(conn, reader.get_fields())

    # yield 0, "count variants..."
    # total_variant = reader.get_variants_count()

    # Insert variants, link them to annotations and samples
    yield 0, "Inserting variants..."
    percent = 0
    for value, message in async_insert_many_variants(conn, reader.get_variants()):

        if reader.file_size:
            percent = reader.read_bytes / reader.file_size * 100.0
        else:
            # Fallback
            # TODO: useless for now because we don't give the total of variants
            # to async_insert_many_variants()
            percent = value
        yield percent, message

    # Create indexes
    yield 99, "Creating indexes..."
    create_indexes(conn)
    yield 100, "Indexes created."

    # session.add(Selection(name="favoris", description="favoris", count = 0))


def async_import_file(conn, filename, project={}):
    """Import filename into SQLite database

    :param conn: sqlite connection
    :param filenaame: variant filename
    :return: yield progression and message
    """
    # Context manager that wraps the given file
    with create_reader(filename) as reader:
        yield from async_import_reader(conn, reader, **project)


def import_file(conn, filename):
    for progress, message in async_import_file(conn, filename):
        #  don't show message
        pass


def import_reader(conn, reader):
    for progress, message in async_import_reader(conn, reader):
        #  don't show message
        pass



# Standard imports
import os
import struct
from contextlib import contextmanager
import binascii
import pathlib

# Custom imports
from .reader import *
import cutevariant.commons as cm
import vcf

LOGGER = cm.logger()


def is_gz_file(filepath):
    """Return a boolean according to the compression state of the file"""
    with open(filepath, "rb") as test_f:
        return binascii.hexlify(test_f.read(2)) == b"1f8b"


def get_uncompressed_size(filename):
    """Get the size of the given compressed file
    This size is stored in the last 4 bytes of the file.
    """
    with open(filename, "rb") as f:
        f.seek(-4, 2)
        return struct.unpack("I", f.read(4))[0]


def detect_vcf_annotation(filename):
    """Return the name of the annotation parser to be used on the given file
    Called: In the importer and in the project wizard to display the detected
    annotations.

    :return: "vep", "snpeff", None
    """
    with open(filename, "r") as file:
        std_reader = vcf.VCFReader(file)
        # print(std_reader.metadata)
        if "VEP" in std_reader.metadata:
            if "CSQ" in std_reader.infos:
                return "vep"

        if "SnpEffVersion" in std_reader.metadata:
            if "ANN" in std_reader.infos:
                return "snpeff"


@contextmanager
def create_reader(filename):
    """Context manager that wraps the given file and return an accurate reader

    A detection of the file type is made as well as a detection of the
    annotations format if required.

    Filetypes and annotations parsers supported:

        - vcf.gz: snpeff, vep
        - vcf: snpeff, vep
        - csv, tsv, txt: vep
    """

    path = pathlib.Path(filename)

    LOGGER.debug(
        "create_reader: PATH suffix %s, is_gz_file: %s",
        path.suffixes,
        is_gz_file(filename),
    )

    if ".vcf" in path.suffixes and ".gz" in path.suffixes:
        # TODO: use is_gz_file() ?
        annotation_detected = detect_vcf_annotation(filename)
        device = open(filename, "rb")
        reader = VcfReader(device, annotation_detected)
        reader.file_size = get_uncompressed_size(filename)
        yield reader
        device.close()
        return

    if ".vcf" in path.suffixes:
        annotation_detected = detect_vcf_annotation(filename)
        device = open(filename, "r")
        reader = VcfReader(device, annotation_detected)
        reader.file_size = os.path.getsize(filename)
        yield reader
        device.close()
        return

    if {".tsv", ".csv", ".txt"} & set(path.suffixes):
        device = open(filename, "r")
        reader = CsvReader(device)
        reader.file_size = os.path.getsize(filename)
        yield reader
        device.close()
        return

    raise Exception("create_reader:: Could not choose parser for this file.")


# class ReaderFactory(object):
#     """
#   Create reader depending file type
#   """

#     def __init__(self):
#         pass

#     @staticmethod
#     def create_reader(filename):

#         if not os.path.isfile(filename):
#             raise FileExistsError()

#         extension = os.path.splitext(filename)[1]

#         if extension == ".csv":
#             print("create csv reader")
#             return CsvReader(open(filename, "r"))

#         if extension == ".vcf":
#             print("create vcf reader")
#             return VcfReader(open(filename, "r"))

#         if extension == ".vcf":
#             print("create vcf reader")
#             return VcfReader(open(filename, "r"))

#         return None



"""List of classes used for settings window

SettingsWidget: Main widget for settings window that instantiate all subwidgets

BaseWidget: Abstract class for settings widgets.
    Subclasses:
        - TranslationSettingsWidget: Allow to choose a language for the interface
        - ProxySettingsWidget: Allow to configure proxy settings for widgets that
        require internet connection
        - StyleSettingsWidget
        - PluginsSettingsWidget
        - DatabaseSettingsWidget
        - VariantSettingsWidget: Allow to add personal templates to search a
        variant in a third-party database

GroupWidget: Handy class to group similar settings widgets in tabs (used by SettingsWidget).
    Used for:
        - TranslationSettingsWidget
        - ProxySettingsWidget
        - StyleSettingsWidget
"""
# Standard imports
import os
import glob
from abc import abstractmethod

# Qt imports
from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *  # QIcon
from PySide2.QtNetwork import *

# Custom imports
import cutevariant.commons as cm
from cutevariant.gui.ficon import FIcon


class BaseWidget(QWidget):
    """Abstract class for settings widgets"""

    def __init__(self):
        super().__init__()
        self.settings = QSettings()

    @abstractmethod
    def save(self):
        """Save the current widget settings in QSettings"""
        raise NotImplementedError(self.__class__.__name__)

    @abstractmethod
    def load(self):
        """Load settings from QSettings"""
        raise NotImplementedError(self.__class__.__name__)


class GroupWidget(QTabWidget):
    """Handy class to group similar settings widgets in tabs"""

    def add_settings_widget(self, widget: BaseWidget):
        self.addTab(widget, widget.windowIcon(), widget.windowTitle())

    def save(self):
        """Call save() method of all widgets in the GroupWidget"""
        [self.widget(index).save() for index in range(self.count())]

    def load(self):
        """Call load() method of all widgets in the GroupWidget"""
        [self.widget(index).load() for index in range(self.count())]


################################################################################
class TranslationSettingsWidget(BaseWidget):
    """Allow to choose a language for the interface"""

    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("Translation"))
        self.setWindowIcon(FIcon(0xF5CA))
        self.locales_combobox = QComboBox()
        mainLayout = QFormLayout()
        mainLayout.addRow(self.tr("&Choose a locale:"), self.locales_combobox)

        self.setLayout(mainLayout)
        # self.locales_combobox.currentTextChanged.connect(self.switchTranslator)

    def save(self):
        """Switch qApp translator with the selected one and save it into config

        .. note:: settings are stored in "ui" group
        .. todo:: Handle the propagation the LanguageChange event
            https://doc.qt.io/Qt-5/qcoreapplication.html#installTranslator
            https://wiki.qt.io/How_to_create_a_multi_language_application
        """

        # Remove the old translator
        # qApp.removeTranslator(translator)

        # Load the new translator

        # Save locale setting
        locale_name = self.locales_combobox.currentText()
        self.settings.setValue("ui/locale", locale_name)
        app_translator = QTranslator(qApp)
        if app_translator.load(locale_name, cm.DIR_TRANSLATIONS):
            qApp.installTranslator(app_translator)

    def load(self):
        """Setup widgets in General settings"""
        self.locales_combobox.clear()
        # Get names of locales based on available files
        available_translations = {
            os.path.basename(os.path.splitext(file)[0]): file
            for file in glob.glob(cm.DIR_TRANSLATIONS + "*.qm")
        }
        # English is the default language
        available_locales = list(available_translations.keys()) + ["en"]
        self.locales_combobox.addItems(available_locales)

        # Display current locale
        locale_name = self.settings.value("ui/locale", "en")
        self.locales_combobox.setCurrentIndex(available_locales.index(locale_name))


class ProxySettingsWidget(BaseWidget):
    """Allow to configure proxy settings for widgets that require internet connection"""

    PROXY_TYPES = {
        "No Proxy": QNetworkProxy.NoProxy,
        "Default": QNetworkProxy.DefaultProxy,
        "Sock5": QNetworkProxy.Socks5Proxy,
        "Http": QNetworkProxy.HttpProxy,
    }

    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("Proxy"))
        self.setWindowIcon(FIcon(0xF484))

        self.combo_box = QComboBox()
        self.host_edit = QLineEdit()
        self.port_edit = QSpinBox()
        self.user_edit = QLineEdit()
        self.pass_edit = QLineEdit()

        # Load proxy type
        for key in self.PROXY_TYPES:
            self.combo_box.addItem(key, self.PROXY_TYPES[key])

        # edit restriction
        self.pass_edit.setEchoMode(QLineEdit.PasswordEchoOnEdit)

        f_layout = QFormLayout()
        f_layout.addRow("Type", self.combo_box)
        f_layout.addRow("Proxy host", self.host_edit)
        f_layout.addRow("Proxy Port", self.port_edit)
        f_layout.addRow("Username", self.user_edit)
        f_layout.addRow("Password", self.pass_edit)

        self.setLayout(f_layout)

    def save(self):
        """Save settings under "proxy" group"""
        settings = QSettings()
        settings.beginGroup("proxy")
        settings.setValue("type", self.combo_box.currentIndex())
        settings.setValue("host", self.host_edit.text())
        settings.setValue("port", self.port_edit.value())
        settings.setValue("username", self.user_edit.text())
        settings.setValue("password", self.user_edit.text())
        settings.endGroup()

    def load(self):
        """Load "proxy" group settings"""
        settings = QSettings()
        settings.beginGroup("proxy")
        self.combo_box.setCurrentIndex(int(settings.value("type", 0)))
        self.host_edit.setText(settings.value("host"))
        self.port_edit.setValue(int(settings.value("port", 0)))
        self.user_edit.setText(settings.value("username"))
        self.pass_edit.setText(settings.value("password"))
        settings.endGroup()


class StyleSettingsWidget(BaseWidget):
    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("Styles"))
        self.setWindowIcon(FIcon(0xF3D8))

    def save(self):
        pass

    def load(self):
        pass


class PluginsSettingsWidget(BaseWidget):
    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("Plugins"))
        self.setWindowIcon(FIcon(0xF3D4))

    def save(self):
        pass

    def load(self):
        pass


class DatabaseSettingsWidget(BaseWidget):
    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("database"))
        self.setWindowIcon(FIcon(0xF1B8))

    def save(self):
        pass

    def load(self):
        pass


class VariantSettingsWidget(BaseWidget):
    """Allow to add personal templates to search a variant in a third-party database

    .. note:: These templates are used in the right click context menu displayed
        on InfoVariantWidget and ViewQueryWidget.
    """

    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("Variants"))
        self.setWindowIcon(FIcon(0xF683))

        self.view = QListWidget()
        self.add_button = QPushButton(self.tr("Add"))
        self.edit_button = QPushButton(self.tr("Edit"))
        self.remove_button = QPushButton(self.tr("Remove"))

        v_layout = QVBoxLayout()
        v_layout.addWidget(self.add_button)
        v_layout.addWidget(self.edit_button)
        v_layout.addStretch()
        v_layout.addWidget(self.remove_button)

        main_layout = QHBoxLayout()
        main_layout.addWidget(self.view)
        main_layout.addLayout(v_layout)

        self.setLayout(main_layout)

        # Settings key
        self.settings_key = "databases_urls/"

        # Signals
        self.add_button.clicked.connect(self.add_url)
        self.edit_button.clicked.connect(self.edit_item)
        self.view.itemDoubleClicked.connect(self.add_url)
        self.remove_button.clicked.connect(self.remove_item)

        # Load built-in databases first
        [
            self.add_list_widget_item(*db_name_url)
            for db_name_url in cm.WEBSITES_URLS.items()
        ]

    def save(self):
        # TODO : save links from settings
        pass

    def load(self):
        """Load databases URLs from settings"""
        # Get all child keys of the group databases_urls
        self.settings.beginGroup(self.settings_key)

        for db_name in self.settings.childKeys():
            # Add the item to the list
            self.add_list_widget_item(db_name, self.settings.value(db_name))

        self.settings.endGroup()

    def add_list_widget_item(self, db_name: str, url: str):
        """Add an item to the QListWidget of the current view"""
        # Key is the name of the database, value is its url
        item = QListWidgetItem(db_name)
        item.setData(Qt.UserRole, url)
        self.view.addItem(item)

    def edit_list_widget_item(self, item: QListWidgetItem, db_name: str, url: str):
        """Modify the given item"""
        item.setText(db_name)
        item.setData(Qt.UserRole, url)

    def add_url(self, item=None):
        """Allow the user to insert and save custom database URL"""
        # Display dialog box to let the user enter it's own url
        dialog = QDialog()
        name = QLineEdit()
        url = QLineEdit()
        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttons.accepted.connect(dialog.accept)
        buttons.rejected.connect(dialog.reject)
        layout = QFormLayout()
        layout.addRow(self.tr("Name"), name)
        layout.addRow(self.tr("Url"), url)
        layout.addWidget(buttons)

        dialog.setLayout(layout)

        if item:
            # Called by itemDoubleClicked or edit_item
            # Fill forms with item data
            name.setText(item.text())
            url.setText(item.data(Qt.UserRole))

        # Also do a minimal check on the data inserted
        if dialog.exec_() == QDialog.Accepted and name.text() and url.text():

            if item:
                # Edit the current item in the list
                self.edit_list_widget_item(item, name.text(), url.text())
            else:
                # Add the item to the list
                self.add_list_widget_item(name.text(), url.text())

            # Save the item in settings
            # (Here to limit the friction with Save all button)
            self.settings.setValue(self.settings_key + name.text(), url.text())

    def edit_item(self):
        """Edit the selected item

        .. note:: This function uses add_url to display the edit window
        """
        # Get selected item
        # Always use the first selected item returned
        self.add_url(self.view.selectedItems()[0])

    def remove_item(self):
        """Remove the selected item

        .. todo:: removeItemWidget() is not functional?
        """
        # Get selected item
        item = self.view.selectedItems()[0]

        # Delete key in settings
        self.settings.remove(self.settings_key + item.text())

        # Delete the item
        self.view.takeItem(self.view.row(item))
        del item  # Is it mandatory in Python ?


class SettingsWidget(QDialog):
    """Main widget for settings window

    Subwidgets are intantiated on panels; a GroupWidget groups similar widgets
    in tabs.
    """

    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("Cutevariant - Settings"))
        self.setWindowIcon(QIcon(cm.DIR_ICONS + "app.png"))
        self.widgets = []

        self.list_widget = QListWidget()
        self.stack_widget = QStackedWidget()
        self.button_box = QDialogButtonBox(
            QDialogButtonBox.SaveAll | QDialogButtonBox.Cancel | QDialogButtonBox.Reset
        )

        self.list_widget.setFixedWidth(200)
        self.list_widget.setIconSize(QSize(32, 32))

        h_layout = QHBoxLayout()
        h_layout.addWidget(self.list_widget)
        h_layout.addWidget(self.stack_widget)

        v_layout = QVBoxLayout()
        v_layout.addLayout(h_layout)
        v_layout.addWidget(self.button_box)
        self.setLayout(v_layout)

        # Instantiate subwidgets on panels
        # Similar widgets for general configuration
        general_settings = GroupWidget()
        general_settings.setWindowTitle(self.tr("General"))
        general_settings.setWindowIcon(FIcon(0xF493))

        general_settings.add_settings_widget(TranslationSettingsWidget())
        general_settings.add_settings_widget(ProxySettingsWidget())
        general_settings.add_settings_widget(StyleSettingsWidget())

        # Specialized widgets on panels
        self.addPanel(general_settings)
        self.addPanel(PluginsSettingsWidget())
        self.addPanel(VariantSettingsWidget())
        self.addPanel(DatabaseSettingsWidget())

        self.resize(800, 400)

        self.button_box.button(QDialogButtonBox.SaveAll).clicked.connect(self.save_all)
        self.button_box.button(QDialogButtonBox.Reset).clicked.connect(self.load_all)
        self.button_box.button(QDialogButtonBox.Cancel).clicked.connect(self.close)

        # Connection events
        self.list_widget.currentRowChanged.connect(self.stack_widget.setCurrentIndex)

        # Load settings
        self.load_all()

    def addPanel(self, widget: BaseWidget):
        """Add a widget on the widow via a QStackedWidget; keep a reference on it
        for later connection/activation"""
        # Used to load/save all widgets on demand
        self.widgets.append(widget)
        # Used for ui positionning and connection events
        self.list_widget.addItem(
            QListWidgetItem(widget.windowIcon(), widget.windowTitle())
        )
        self.stack_widget.addWidget(widget)

    def save_all(self):
        """Call save() method of all widgets"""
        [widget.save() for widget in self.widgets]

    def load_all(self):
        """Call load() method of all widgets"""
        [widget.load() for widget in self.widgets]



"""Proof of concept for the VQL DSL.

See test_vql.py for usage and features.

"""
import textx
from pkg_resources import resource_string


def model_class(name: str, bases: tuple, attrs: dict) -> type:
    """Metaclass to automatically build the __init__ to get the properties,
    and register the class for metamodel
    """
    if "__init__" not in attrs:

        def __init__(self, *args, **kwargs):
            for field, value in kwargs.items():
                setattr(self, field, value)

        attrs["__init__"] = __init__
    cls = type(name, bases, attrs)
    model_class.classes.append(cls)
    return cls


model_class.classes = []


class VQLSyntaxError(ValueError):
    def __init__(self, message, col=None, *args, **kwargs):
        super().__init__(message, col, *args, **kwargs)
        self.message = message
        self.col = col


# ============ Error handle ==================================
def error_message_from_err(
    err: textx.exceptions.TextXSyntaxError, raw_vql: str
) -> (str, int):
    """Return human-readable information and index in raw_sql query
    about the given exception"""
    # print(err)
    # print(dir(err))
    # print(err.args)
    # print(err.err_type)
    # print(err.line, err.col)
    # print(err.message)
    # print(err.filename)
    # print(err.expected_rules)
    if "'SELECT'" in err.message:  # was awaiting for a SELECT clause
        return "no SELECT clause", -1
    if err.message.endswith("=> 's,ref FROM*'."):
        return "empty 'FROM' clause", err.col
    if (
        ",*," in err.message
        and len(err.expected_rules) == 1
        and type(err.expected_rules[0]).__name__ == "RegExMatch"
    ):
        return "invalid empty identifier in SELECT clause", err.col
    if "Expected INT " in err.message and len(err.expected_rules) == 3:
        return "invalid value in WHERE clause", err.col
    if "Expected '==|>=|<=|!=" in err.message and len(err.expected_rules) == 1:
        return "invalid operator in WHERE clause", err.col

    raise err  # error not handled. Just raise it


# ============ Different class to map with VQL model


class FilterTerm(metaclass=model_class):
    @property
    def value(self):
        field = self.field.value if hasattr(self.field, "value") else self.field
        val = self.val.value if hasattr(self.val, "value") else self.val

        # escape if quoted
        if isinstance(val, str):
            val = f"'{val}'"

        return {"field": field, "operator": self.op, "value": val}


class FilterExpression(metaclass=model_class):
    @property
    def value(self):
        out = []
        key = "AND"  #  By default
        for i in self.op:
            if isinstance(i, str):
                if i in ("AND", "OR"):
                    key = i
                else:
                    out.append(i)
            else:
                out.append(i.value)
        return {key: out}


class FilterOperand(metaclass=model_class):
    @property
    def value(self):
        return self.op.value


class Function(metaclass=model_class):
    @property
    def value(self):
        return (self.func, self.arg, "gt")


class Tuple(metaclass=model_class):
    @property
    def value(self):
        return tuple(self.items)


class SelectCmd(metaclass=model_class):
    @property
    def value(self):
        output = {
            "cmd": "select_cmd",
            "columns": [
                col.value if hasattr(col, "value") else col for col in self.columns
            ],
            "source": self.source,
        }

        if self.filter:
            output["filter"] = self.filter.value

        return output


class CreateCmd(metaclass=model_class):
    @property
    def value(self):
        return {
            "cmd": "create_cmd",
            "source": self.source,
            "fitler": self.filter.value if self.filter else None,
        }


# class SetCmd(metaclass=model_class):
#     @property
#     def value(self):
#         return {
#         "cmd": "set_cmd",
#         "source": self.source,
#         "fitler": self.filter.value if self.filter else None
#         }


METAMODEL = textx.metamodel_from_str(
    resource_string(__name__, "vql.tx").decode(),  # grammar extraction from vql.tx
    classes=model_class.classes,
    debug=False,
    ignore_case=True,
)


def execute_vql(raw_vql: str) -> list:
    """Execute multiline VQL statement separated by ";"

    :return: yield 1 dictionnary per command
        .. example :: {'cmd': 'select_cmd', 'columns': ['chr','pos'], 'source':'variants', 'filter': 'None'}
    """
    try:
        raw_model = METAMODEL.model_from_str(raw_vql)
    except textx.exceptions.TextXSyntaxError as err:
        raise VQLSyntaxError(*error_message_from_err(err, raw_vql))

    yield from (command.value for command in raw_model.commands)


def model_from_string(raw_vql: str) -> dict:
    """Obsolete : retro compatibility"""
    return next(execute_vql(raw_vql))



from PySide2.QtCore import *
from PySide2.QtWidgets import *
from PySide2.QtGui import *

from cutevariant.core.query import Query
from cutevariant.gui.ficon import FIcon
import re


class OmniBar(QLineEdit):

    changed = Signal()

    def __init__(self):
        super().__init__()
        self.setMaximumWidth(400)
        self.setPlaceholderText(qApp.tr("chr3:100-100000"))
        self.action = self.addAction(FIcon(0xF349, "black"), QLineEdit.TrailingPosition)

        self.returnPressed.connect(self.changed)
        self._query = None

    def to_coordinate(self):

        match = re.search(r"(chr\d\d?):(\d+)-(\d+)", self.text())

        if match:
            return (match[1], int(match[2]), int(match[3]))
        else:
            return None

    @property
    def query(self):

        coord = self.to_coordinate()
        if coord:
            self._query.filter = {
                "AND": [
                    {"field": "chr", "operator": "=", "value": coord[0]},
                    {"field": "pos", "operator": ">", "value": coord[1]},
                    {"field": "pos", "operator": "<", "value": coord[2]},
                ]
            }

        return self._query

    @query.setter
    def query(self, query: Query):
        self._query = query



from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *
from enum import Enum

from .plugin import QueryPluginWidget
from cutevariant.core import Query
from cutevariant.gui.ficon import FIcon


class FilterType(Enum):
    LOGIC = 1
    CONDITION = 2


class LogicItem(QStandardItem):

    _LOGIC_ICONS = {"AND": 0xF8E0, "OR": 0xF8E4}

    def __init__(self, logic_type="AND"):
        """
        Create a logic Item : OR / AND  
        """
        super().__init__()
        self.logic_type = logic_type
        self.setEditable(False)
        self.set(logic_type)

    def set(self, logic_type):
        self.setIcon(FIcon(LogicItem._LOGIC_ICONS[self.logic_type]))
        self.setText(logic_type)


class FieldItem(QStandardItem):
    def __init__(self, name, operator, value):
        super().__init__()
        self.setEditable(False)
        self.set(name, operator, value)

    def set(self, name, operator, value):
        self.name = name
        self.operator = operator
        self.value = value
        self.setText(f"{self.name}  {self.operator}  {self.value}")


class FilterQueryModel(QStandardItemModel):
    def __init__(self):
        super().__init__()
        self._query = None

    @property
    def query(self) -> Query:
        if self.rowCount() != 0:
            self._query.filter = self.fromItem(self.item(0))
        return self._query

    @query.setter
    def query(self, query: Query):
        self._query = query

    def load(self):
        self.clear()
        if self._query.filter:
            self.appendRow(self.toItem(self._query.filter))

    def toItem(self, data: dict) -> QStandardItem:
        """ recursive function to load item in tree from data """
        if len(data) == 1:  #  logic item
            operator = list(data.keys())[0]
            item = LogicItem(operator)
            [item.appendRow(self.toItem(k)) for k in data[operator]]
            return item
        else:  # condition item
            item = FieldItem(data["field"], data["operator"], data["value"])
            return item

    def fromItem(self, item: QStandardItem) -> dict:
        """ recursive fonction to get items from tree """
        if isinstance(item, LogicItem):
            # Return dict with operator as key and item as value
            operator_data = [
                self.fromItem(item.child(i)) for i in range(item.rowCount())
            ]
            return {item.logic_type: operator_data}
        else:
            return {"field": item.name, "operator": item.operator, "value": item.value}


class FilterEditDialog(QDialog):
    def __init__(self, item: QStandardItem):
        super().__init__()
        self.field = QLineEdit()
        self.operator = QLineEdit()
        self.value = QLineEdit()
        self.item = item

        form_layout = QFormLayout()

        form_layout.addRow("field", self.field)
        form_layout.addRow("operator", self.operator)
        form_layout.addRow("value", self.value)

        layout = QVBoxLayout()
        buttonBox = QDialogButtonBox(QDialogButtonBox.Save | QDialogButtonBox.Cancel)
        buttonBox.accepted.connect(self.save)
        buttonBox.rejected.connect(self.close)

        layout.addLayout(form_layout)
        layout.addWidget(buttonBox)
        self.setLayout(layout)

        self.load()

    def load(self):
        self.field.setText(self.item.data(FilterQueryModel.FieldRole))
        self.operator.setText(self.item.data(FilterQueryModel.OperatorRole))
        self.value.setText(self.item.data(FilterQueryModel.ValueRole))

    def save(self):
        self.item.setText(self.field.text() + self.operator.text() + self.value.text())
        self.item.setData("field", FilterQueryModel.TypeRole)
        self.item.setData(self.field.text(), FilterQueryModel.FieldRole)
        self.item.setData(self.operator.text(), FilterQueryModel.OperatorRole)
        self.item.setData(self.value.text(), FilterQueryModel.ValueRole)


class FilterQueryWidget(QueryPluginWidget):
    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("Filter"))
        self.view = QTreeView()
        self.model = FilterQueryModel()
        self.view.setModel(self.model)

        layout = QVBoxLayout()
        layout.addWidget(self.view)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

        # self.model.itemChanged.connect(self.changed)
        self.view.doubleClicked.connect(self.edit)

    def on_init_query(self):
        """ Overrided """
        self.model.query = self.query

    def on_change_query(self):
        """ override methods """
        self.model.load()

    def edit(self, index):
        dialog = FilterEditDialog(self.model.itemFromIndex(index))

        if dialog.exec_():
            dialog.save()

    def contextMenuEvent(self, event):
        """ override methode """

        menu = QMenu(self)

        pos = self.view.viewport().mapFromGlobal(event.globalPos())
        index = self.view.indexAt(pos)

        if not index.isValid():
            logic_action = menu.addAction(self.tr("add logic"))

        else:
            logic_action = menu.addAction(self.tr("add logic"))
            item_action = menu.addAction(self.tr("add condition"))
            menu.addSeparator()
            rem_action = menu.addAction(self.tr("Remove"))

        menu.exec_(event.globalPos())



"""Expose AboutCutevariant class to show information about the project"""
# Standard imports
from pkg_resources import resource_filename

# Qt imports
from PySide2.QtCore import Qt, QRect, QUrl
from PySide2.QtWidgets import (
    QDialog,
    QTabWidget,
    QLabel,
    QDialogButtonBox,
    QVBoxLayout,
    QPlainTextEdit,
    QFrame,
)
from PySide2.QtGui import (
    QIcon,
    QPixmap,
    QPainter,
    QBrush,
    QFont,
    QPen,
    QColor,
    QFontMetrics,
    QDesktopServices,
)

# Custom imports
from cutevariant.gui.ficon import FIcon
from cutevariant import commons as cm
from cutevariant import __version__


class AboutCutevariant(QDialog):
    """Display a dialog window with information about the project

    The window is mainly about the project license, authors, history of changes.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle(self.tr("About Cutevariant"))
        self.setWindowIcon(QIcon(cm.DIR_ICONS + "app.png"))

        self.tab_widget = QTabWidget()
        self.header_lbl = QLabel()
        self.button_box = QDialogButtonBox()

        github_button = self.button_box.addButton("GitHub", QDialogButtonBox.HelpRole)
        twitter_button = self.button_box.addButton("Twitter", QDialogButtonBox.HelpRole)

        self.button_box.addButton(QDialogButtonBox.Ok)
        github_button.setIcon(FIcon(0xF2A4))
        twitter_button.setIcon(FIcon(0xF546))

        vLayout = QVBoxLayout()
        vLayout.addWidget(self.header_lbl)
        vLayout.addWidget(self.tab_widget)
        vLayout.addWidget(self.button_box)

        self.setLayout(vLayout)

        self.addTab("LICENSE")
        # self.addTab("AUTHORS")
        # self.addTab("CREDITS")
        # self.addTab("CHANGELOG")

        self.drawHeader()

        # Connexions
        self.button_box.button(QDialogButtonBox.Ok).clicked.connect(self.close)
        github_button.clicked.connect(self.open_github_page)
        twitter_button.clicked.connect(self.open_twitter_page)

        self.resize(600, 400)

    def addTab(self, filename):
        """Read the given text file and load it in a new tab"""
        # Get file at the project's root directory
        filepath = resource_filename(__name__, "../../" + filename)

        text_edit = QPlainTextEdit()
        text_edit.setFrameShape(QFrame.NoFrame)
        text_edit.setReadOnly(True)

        text = ""
        if filename == "LICENSE":
            # Add license header
            text = self.tr(
                "Copyright (C) 2018-2019  Labsquare.org\n\n"
                "This program is distributed under the terms of the GNU "
                "General Public License v3.\n\n"
            )

        with open(filepath, "r") as f_d:
            text_edit.setPlainText(text + f_d.read())
            self.tab_widget.addTab(text_edit, filename)

    def drawHeader(self):
        """Draw logo/copyright in the header"""
        pHeight = 90
        pMargin = 15
        icon_path = cm.DIR_ICONS + "app.png"

        self.header_lbl.setMinimumHeight(pHeight)
        self.header_lbl.setFrameShape(QFrame.StyledPanel)
        self.header_lbl.setContentsMargins(0, 0, 0, 0)

        pixmap = QPixmap(450, pHeight)
        pixmap.fill(Qt.transparent)

        iconY = (pHeight - 64) / 2
        logoRect = QRect(pMargin, iconY, 64, 64)

        painter = QPainter(pixmap)
        painter.setBrush(QBrush(Qt.red))
        painter.drawPixmap(
            logoRect,
            QPixmap(icon_path).scaled(
                logoRect.width(),
                logoRect.height(),
                Qt.KeepAspectRatio,
                Qt.SmoothTransformation,
            ),
        )

        titleRect = QRect(logoRect.right() + 10, iconY, 200, pHeight)

        font = QFont()
        font.setBold(True)
        font.setPixelSize(16)
        painter.setFont(font)
        painter.setPen(QPen(qApp.palette().text().color()))
        painter.drawText(titleRect, Qt.AlignTop, "Cutevariant")

        font_metrics = QFontMetrics(font)
        font.setBold(False)
        font.setPixelSize(12)
        painter.setFont(font)
        painter.setPen(QPen(Qt.darkGray))
        titleRect.setY(titleRect.y() + font_metrics.height())

        painter.drawText(
            titleRect,
            Qt.AlignTop,
            f"Version %s\nGPL3 Copyright (C) 2018-2019\nLabsquare.org" % __version__,
        )

        self.header_lbl.setPixmap(pixmap)

        # Painting is finished !
        # Avoid Segfault:
        # QPaintDevice: Cannot destroy paint device that is being painted
        painter.end()

    def open_github_page(self):
        """Open the project page on GitHub"""
        QDesktopServices.openUrl(QUrl("https://github.com/labsquare/cutevariant"))

    def open_twitter_page(self):
        """Open the labsquare page on Twitter"""
        QDesktopServices.openUrl(QUrl("https://twitter.com/labsquare"))



"""Module bringing together all the SQL related functions.

- Misc functions
- Selections functions
- Fields functions
- Operations on sets of variants
- Annotations functions
- Variants functions
- Samples functions
"""

# Standard imports
import sqlite3
import sys
from collections import defaultdict
from pkg_resources import parse_version

# Custom imports
import cutevariant.commons as cm

LOGGER = cm.logger()


## ================ Misc functions =============================================


def get_sql_connexion(filepath):
    """Open a SQLite database and return the connexion object"""
    connexion = sqlite3.connect(filepath)
    # Activate Foreign keys
    connexion.execute("PRAGMA foreign_keys = ON")
    connexion.row_factory = sqlite3.Row
    foreign_keys_status = connexion.execute("PRAGMA foreign_keys").fetchone()[0]
    LOGGER.debug("get_sql_connexion:: foreign_keys state: %s", foreign_keys_status)
    assert foreign_keys_status == 1, "Foreign keys can't be activated :("
    return connexion


def drop_table(conn, table_name):
    """Drop the given table"""
    cursor = conn.cursor()
    cursor.execute(f"DROP TABLE IF EXISTS {table_name}")
    conn.commit()


def create_project(conn, name: str, reference: str):
    """Create the table "projects" and insert project name and reference genome

    :param conn: sqlite3.connect
    :param name: Project's name
    :param reference: Reference genome
    """
    cursor = conn.cursor()
    cursor.execute(
        """CREATE TABLE projects (id INTEGER PRIMARY KEY, name TEXT, reference TEXT)"""
    )
    cursor.execute(
        """INSERT INTO projects (name, reference) VALUES (?, ?)""", (name, reference)
    )
    conn.commit()


def get_columns(conn, table_name):
    """Return the list of columns for the given table

    .. note:: used by async_insert_many_variants()
        to build queries with placeholders

    :param conn: sqlite3.connect
    :param table_name: Table for which columns will be returned.
    """
    # Get columns description from table_info
    # ((0, 'chr', 'str', 0, None, 1), ...
    return [
        c[1] for c in conn.execute(f"pragma table_info({table_name})") if c[1] != "id"
    ]


def create_indexes(conn):
    """Create extra indexes on tables

    .. note:: This function must be called after batch insertions.
    .. note:: You should use this function instead of individual functions.
    """
    create_variants_indexes(conn)
    create_selections_indexes(conn)

    try:
        # Some databases have not annotations table
        create_annotations_indexes(conn)
    except sqlite3.OperationalError as e:
        LOGGER.debug("create_indexes:: sqlite3.%s: %s", e.__class__.__name__, str(e))


## ================ Selections functions =======================================


def create_table_selections(conn):
    """Create the table "selections" and association table "selection_has_variant"

    This table stores the queries saved by the user:
        - name: name of the set of variants
        - count: number of variants concerned by this set
        - query: the SQL query which generated the set

    :param conn: sqlite3.connect
    """
    cursor = conn.cursor()
    # selection_id is an alias on internal autoincremented 'rowid'
    cursor.execute(
        """CREATE TABLE selections (
        id INTEGER PRIMARY KEY ASC,
        name TEXT, count INTEGER, query TEXT
        )"""
    )

    # Association table: do not use useless rowid column
    cursor.execute(
        """CREATE TABLE selection_has_variant (
        variant_id INTEGER NOT NULL,
        selection_id INTEGER NOT NULL,
        PRIMARY KEY (variant_id, selection_id),
        FOREIGN KEY (selection_id) REFERENCES selections (id)
          ON DELETE CASCADE
          ON UPDATE NO ACTION
        ) WITHOUT ROWID"""
    )
    conn.commit()


def create_selections_indexes(conn):
    """Create indexes on the "selections" table

    .. note:: This function should be called after batch insertions.

    .. note:: This function ensures the unicity of selections names.
    """
    conn.execute("""CREATE UNIQUE INDEX idx_selections ON selections (name)""")


def create_selection_has_variant_indexes(conn):
    """Create indexes on "selection_has_variant" table

    .. note:: This function is called by:
        - create_selections_indexes
        - insert_selection to rebuild index
    """
    # For joints between selections and variants tables
    conn.execute(
        """CREATE INDEX idx_selection_has_variant ON selection_has_variant (selection_id)"""
    )


def insert_selection(conn, query: str, name="no_name", count=0):
    """Insert one selection record

    TODO: retirer le group by inutile

    .. warning:: This function does a commit !

    :param conn: sqlite3 connection OR cursor
        If connection: commit is made.
        If cursor: commit is not made.
    :param name: name of the selection
    :param count: precompute variant count
    :param query: Sql variant query selection
    :type conn: <sqlite3.Connection> or <sqlite3.Cursor>
    :return: rowid of the new selection inserted.
    :rtype: <int>

    .. seealso:: create_selection_from_sql
    """
    # conn can be a cursor or a connection here...
    cursor = conn.cursor() if isinstance(conn, sqlite3.Connection) else conn

    cursor.execute(
        """INSERT INTO selections (name, count, query) VALUES (?,?,?)""",
        (name, count, query),
    )
    if isinstance(conn, sqlite3.Connection):
        # Commit only if connection is given. => avoid not consistent DB
        conn.commit()
    return cursor.lastrowid


def create_selection_from_sql(
    conn, query: str, name: str, count=None, from_selection=False
):
    """Create a selection record from sql variant query

    :param name : name of the selection
    :param query: sql variant query
    :param by: can be : 'site' for (chr,pos)  or 'variant' for (chr,pos,ref,alt)
    :return: The id of the new selection. None in case of error.
    :rtype: <int> or None
    """
    cursor = conn.cursor()

    # Compute query count
    #  TODO : this can take a while .... need to compute only one from elsewhere
    if not count:
        count = cursor.execute(f"SELECT COUNT(*) FROM ({query})").fetchone()[0]

    # Create selection
    selection_id = insert_selection(cursor, query, name=name, count=count)

    # DROP indexes
    # For joints between selections and variants tables
    try:
        cursor.execute("""DROP INDEX idx_selection_has_variant""")
    except sqlite3.OperationalError:
        pass

    # Insert into selection_has_variant table
    # PS: We use DISTINCT keyword to statisfy the unicity constraint on
    # (variant_id, selection_id) of "selection_has_variant" table.
    # TODO: is DISTINCT useful here? How a variant could be associated several
    # times with an association?
    if from_selection:
        # Optimized only for the creation of a selection from set operations
        # variant_id is the only useful column here
        q = f"""
        INSERT INTO selection_has_variant
        SELECT DISTINCT variant_id, {selection_id} FROM ({query})
        """
    else:
        # Fallback
        # Used when creating a selection from a VQL query in the UI
        # Default behavior => a variant is based on chr,pos,ref,alt
        q = f"""
        INSERT INTO selection_has_variant
        SELECT DISTINCT id, {selection_id} FROM ({query})
        """

    cursor.execute(q)

    # REBUILD INDEXES
    # For joints between selections and variants tables
    create_selection_has_variant_indexes(cursor)

    conn.commit()
    if cursor.rowcount:
        return cursor.lastrowid
    return None


def get_selections(conn):
    """Get selections in "selections" table

    :return: Generator of dictionnaries with as many keys as there are columns
        in the table.
        Dictionnary of all attributes of the table.
            :Example: {"id": ..., "name": ..., "count": ..., "query": ...}
    :rtype: <generator <dict>>
    """
    conn.row_factory = sqlite3.Row
    return (dict(data) for data in conn.execute("""SELECT * FROM selections"""))


def delete_selection(conn, selection_id: int):
    """Delete the selection with the given id in the "selections" table

    :return: Number of rows deleted
    :rtype: <int>
    """
    # ON CASCADE deletion
    cursor = conn.cursor()
    cursor.execute("DELETE FROM selections WHERE rowid = ?", (selection_id,))
    conn.commit()
    return cursor.rowcount


def edit_selection(conn, selection: dict):
    """Update the name and count of a selection with the given id

    :return: Number of rows deleted
    :rtype: <int>
    """
    cursor = conn.cursor()
    conn.execute(
        "UPDATE selections SET name=:name, count=:count WHERE id = :id", selection
    )
    conn.commit()
    return cursor.rowcount


## ================ Operations on sets of variants =============================


def get_query_columns(mode="variant"):
    """(DEPRECATED FOR NOW, NOT USED)
    Handy func to get columns to be queried according to the group by argument

    .. note:: Used by intersect_variants, union_variants, subtract_variants
        in order to avoid code duplication.
    """
    if mode == "site":
        return "chr,pos"

    if mode == "variant":
        # Not used
        return "variant_id"

    raise NotImplementedError


def intersect_variants(query1, query2, **kwargs):
    """Get the variants obtained by the intersection of 2 queries"""
    return f"""{query1} INTERSECT {query2}"""


def union_variants(query1, query2, **kwargs):
    """Get the variants obtained by the union of 2 queries"""
    return f"""{query1} UNION {query2}"""


def subtract_variants(query1, query2, **kwargs):
    """Get the variants obtained by the difference of 2 queries"""
    return f"""{query1} EXCEPT {query2}"""


## ================ Fields functions ===========================================


def create_table_fields(conn):
    """Create the table "fields"

    This table contain fields. A field is a column with its description and type;
    it can be choose by a user to build a Query
    Fields are the columns of the tables: variants, annotations and sample_has_variant.
    Fields are extracted from reader objects and are dynamically constructed.

    variants:
    Chr,pos,ref,alt, filter, qual, dp, af, etc.

    annotations:
    Gene, transcrit, etc.

    sample_has_variant:
    Genotype

    :param conn: sqlite3.connect
    """
    cursor = conn.cursor()
    cursor.execute(
        """CREATE TABLE fields
        (id INTEGER PRIMARY KEY, name TEXT, category TEXT, type TEXT, description TEXT)
        """
    )
    conn.commit()


def insert_field(
    conn, name="no_name", category="variants", type="text", description=str()
):
    """Insert one field record (NOT USED)

    :param conn: sqlite3.connect
    :key name: field name
    :key category: category field name.
        .. warning:: The default is "variants". Don't use sample as category name
    :key type: sqlite type which can be: INTEGER, REAL, TEXT
        .. todo:: Check this argument...
    :key description: Text that describes the field (showed to the user).
    :return: Last row id
    :rtype: <int>
    """
    cursor = conn.cursor()
    cursor.execute(
        """
        INSERT INTO fields VALUES (?, ?, ?, ?)
        """,
        (name, category, type, description),
    )
    conn.commit()
    return cursor.lastrowid


def insert_many_fields(conn, data: list):
    """Insert multiple fields into "fields" table using one commit

    :param conn: sqlite3.connect
    :param data: list of field dictionnary

    :Examples:

        insert_many_fields(conn, [{name:"sacha", category:"variant", count: 0, description="a description"}])
        insert_many_fields(conn, reader.get_fields())

    .. seealso:: insert_field, abstractreader
    """
    cursor = conn.cursor()
    cursor.executemany(
        """
        INSERT INTO fields (name,category,type,description)
        VALUES (:name,:category,:type,:description)
        """,
        data,
    )
    conn.commit()


def get_fields(conn):
    """Get fields as list of dictionnary

    .. seealso:: insert_many_fields

    :param conn: sqlite3.connect
    :return: Generator of dictionnaries with as many keys as there are columns
        in the table.
    :rtype: <generator <dict>>
    """
    conn.row_factory = sqlite3.Row
    return (dict(data) for data in conn.execute("""SELECT * FROM fields"""))


## ================ Annotations functions ======================================


def create_table_annotations(conn, fields):
    """Create "annotations" table which contains dynamics fields

    :param fields: Generator of SQL fields.
        :Example of fields:
            ('allele str NULL', 'consequence str NULL', ...)
    :type fields: <generator>
    """
    schema = ",".join([f'{field["name"]} {field["type"]}' for field in fields])

    if not schema:
        #  Create minimum annotation table... Can be use later for dynamic annotation.
        # TODO : we may want to fix annotation fields .
        schema = "gene TEXT, transcript TEXT"
        LOGGER.debug(
            "create_table_annotations:: No annotation fields detected! => Fallback"
        )
        # return

    cursor = conn.cursor()
    # TODO: no primary key/unique index for this table?
    cursor.execute(
        f"""CREATE TABLE annotations (variant_id INTEGER NOT NULL, {schema})"""
    )
    conn.commit()


def create_annotations_indexes(conn):
    """Create indexes on the "annotations" table

    .. warning: This function must be called after batch insertions.

    :Example:
        SELECT *, group_concat(annotations.rowid) FROM variants
        LEFT JOIN annotations ON variants.rowid = annotations.variant_id
        WHERE pos = 0
        GROUP BY chr,pos
        LIMIT 100
    """

    # Allow search on variant_id
    conn.execute("""CREATE INDEX idx_annotations ON annotations (variant_id)""")


## ================ Variants functions =========================================


def create_table_variants(conn, fields):
    """Create "variants" and "sample_has_variant" tables which contains dynamics fields

    :Example:

        fields = get_fields()
        create_table_variants(conn, fields)

    .. seealso:: get_fields

    .. note:: "gt" field in "sample_has_variant" = Patient's genotype.
        - Patient without variant: gt = 0: Wild homozygote
        - Patient with variant in the heterozygote state: gt = -1: Heterozygote
        - Patient with variant in the homozygote state: gt = 2: Homozygote

        :Example of VQL query:
            SELECT chr, pos, genotype("pierre") FROM variants

    :param conn: sqlite3.connect
    :param fields: list of field dictionnary.
    """
    cursor = conn.cursor()

    # Primary key MUST NOT have NULL fields !
    # PRIMARY KEY should always imply NOT NULL.
    # Unfortunately, due to a bug in some early versions, this is not the case in SQLite.
    # For the purposes of UNIQUE constraints, NULL values are considered distinct
    # from all other values, including other NULLs.
    schema = ",".join(
        [
            f'`{field["name"]}` {field["type"]} {field.get("constraint", "")}'
            for field in fields
            if field["name"]
        ]
    )

    LOGGER.debug("create_table_variants:: schema: %s", schema)
    # Unicity constraint or NOT NULL fields (Cf VcfReader, FakeReader, etc.)
    # NOTE: specify the constraint in CREATE TABLE generates a lighter DB than
    # a separated index... Don't know why.
    cursor.execute(
        f"""CREATE TABLE variants (id INTEGER PRIMARY KEY, {schema},
        UNIQUE (chr,pos,ref,alt))"""
    )
    # cursor.execute(f"""CREATE UNIQUE INDEX idx_variants_unicity ON variants (chr,pos,ref,alt)""")

    # Association table: do not use useless rowid column
    cursor.execute(
        """CREATE TABLE sample_has_variant (
        sample_id INTEGER NOT NULL,
        variant_id INTEGER NOT NULL,
        gt INTEGER DEFAULT -1,
        PRIMARY KEY (sample_id, variant_id),
        FOREIGN KEY (sample_id) REFERENCES samples (id)
          ON DELETE CASCADE
          ON UPDATE NO ACTION
        ) WITHOUT ROWID"""
    )
    conn.commit()


def create_variants_indexes(conn):
    """Create indexes on the "variants" table

    .. warning: This function must be called after batch insertions.

    :Example:
        SELECT *, group_concat(annotations.rowid) FROM variants
        LEFT JOIN annotations ON variants.rowid = annotations.variant_id
        WHERE pos = 0
        GROUP BY chr,pos
        LIMIT 100
    """

    # Complementary index of the primary key (sample_id, variant_id)
    conn.execute(
        """CREATE INDEX idx_sample_has_variant ON sample_has_variant (variant_id)"""
    )

    conn.execute("""CREATE INDEX idx_variants_pos ON variants (pos)""")
    conn.execute("""CREATE INDEX idx_variants_ref_alt ON variants (ref, alt)""")


def get_one_variant(conn, id: int):
    """Get the variant with the given id"""
    # Use row_factory here
    conn.row_factory = sqlite3.Row
    # Cast sqlite3.Row object to dict because later, we use items() method.
    return dict(
        conn.execute(f"""SELECT * FROM variants WHERE variants.id = {id}""").fetchone()
    )


def get_variants_count(conn):
    """Get the number of variants in the "variants" table"""
    return conn.execute("""SELECT COUNT(*) FROM variants""").fetchone()[0]


def async_insert_many_variants(conn, data, total_variant_count=None, yield_every=3000):
    """Insert many variants from data into variants table

    :param conn: sqlite3.connect
    :param data: list of variant dictionnary which contains same number of key than fields numbers.
    :param variant_count: total variant count, to compute progression
    :return: Yield a tuple with progression and message.
        Progression is 0 if total_variant_count is not set.
    :rtype: <generator <tuple <int>, <str>>


    :Example:

        insert_many_variant(conn, [{chr:"chr1", pos:24234, alt:"A","ref":T }])
        insert_many_variant(conn, reader.get_variants())

    .. warning:: Using reader, this can take a while
    .. todo:: with large dataset, need to cache import
    .. todo:: handle insertion errors...
    .. seealso:: abstractreader

    .. warning:: About using INSERT OR IGNORE:
        INSERT OR IGNORE avoids errors:

            - Upon insertion of a duplicate key where the column must contain
            a PRIMARY KEY or UNIQUE constraint
            - Upon insertion of NULL value where the column has
            a NOT NULL constraint.
        => This is not recommended
    """

    def build_columns_and_placeholders(table_name):
        """Build a tuple of columns and formatted placeholders for INSERT queries
        """
        # Get columns description from the given table
        cols = get_columns(conn, table_name)
        # Build dynamic insert query
        # INSERT INTO variant qcol1, qcol2.... VALUES :qcol1, :qcol2 ....
        tb_cols = ",".join([f"`{col}`" for col in cols])
        tb_places = ",".join([f":{place}" for place in cols])
        return tb_cols, tb_places

    # TODO: Can we avoid this step ? This function should receive columns names
    # because all the tables were created before...
    # Build placeholders
    var_cols, var_places = build_columns_and_placeholders("variants")
    ann_cols, ann_places = build_columns_and_placeholders("annotations")

    # Get samples with samples names as keys and sqlite rowid as values
    # => used as a mapping for samples ids
    samples_id_mapping = {
        name: rowid for name, rowid in conn.execute("SELECT name, id FROM samples")
    }
    samples_names = samples_id_mapping.keys()

    # Check SQLite version and build insertion queries for variants
    # Old version doesn't support ON CONFLICT ..target.. DO ... statements
    # to handle violation of unicity constraint.
    old_sqlite_version = parse_version(sqlite3.sqlite_version) < parse_version("3.24.0")

    if old_sqlite_version:
        LOGGER.warning(
            "async_insert_many_variants:: Old SQLite version: %s"
            " - Fallback to ignore errors!",
            sqlite3.sqlite_version,
        )
        # /!\ This syntax is SQLite specific
        # /!\ We mask all errors here !
        variant_insert_query = f"""INSERT OR IGNORE INTO variants ({var_cols})
                VALUES ({var_places})"""

    else:
        # Handle conflicts on the primary key
        variant_insert_query = f"""INSERT INTO variants ({var_cols})
                VALUES ({var_places})
                ON CONFLICT (chr,pos,ref,alt) DO NOTHING"""

    # Insertion - Begin transaction
    cursor = conn.cursor()

    # Loop over variants
    errors = 0
    progress = 0
    for variant_count, variant in enumerate(data, 1):

        # Insert current variant
        # Use default dict to handle missing values
        LOGGER.debug(variant_insert_query)
        print(variant_insert_query)

        cursor.execute(variant_insert_query, defaultdict(str, variant))

        # If the row is not inserted we skip this erroneous variant
        # and the data that goes with
        if cursor.rowcount == 0:
            LOGGER.error(
                "async_insert_many_variants:: The following variant "
                "contains erroneous data; most of the time it is a "
                "duplication of the primary key: (chr,pos,ref,alt). "
                "Please check your data; this variant and its attached "
                "data will not be inserted!\n%s",
                variant,
            )
            errors += 1
            continue

        # Get variant rowid
        variant_id = cursor.lastrowid

        # If variant has annotation data, insert record into "annotations" table
        # One-to-many relationships
        if "annotations" in variant:
            # print("VAR annotations:", variant["annotations"])

            # [{'allele': 'T', 'consequence': 'intergenic_region', 'impact': 'MODIFIER', ...}]
            # The aim is to execute all insertions through executemany()
            # We remove the placeholder :variant_id from places,
            # and fix it's value.
            # TODO: handle missing values;
            # Les dict de variant["annotations"] contiennent a priori déjà
            # tous les champs requis (mais vides) car certaines annotations
            # ont des données manquantes.
            # A t'on l'assurance de cela ?
            # Dans ce cas pourquoi doit-on bricoler le variant lui-meme avec un
            # defaultdict(str,variant)) ? Les variants n'ont pas leurs champs par def ?
            temp_ann_places = ann_places.replace(":variant_id,", "")
            cursor.executemany(
                f"""INSERT INTO annotations ({ann_cols})
                VALUES ({variant_id}, {temp_ann_places})""",
                variant["annotations"],
            )

        # If variant has sample data, insert record into "sample_has_variant" table
        # Many-to-many relationships
        if "samples" in variant:
            # print("VAR samples:", variant["samples"])
            # [{'name': 'NORMAL', 'gt': 1, 'AD': '64,0', 'AF': 0.0, ...}]
            # Insertion only if the current sample name is in samples_names
            # (authorized sample names already in the database)
            # TODO: is this test usefull since samples that are in the database
            # have been inserted from the same source file (or it is not the case ?) ?
            # Retrieve the id of the sample to build the association in
            # "sample_has_variant" table carrying the data "gt" (genotype)
            g = (
                (samples_id_mapping[sample["name"]], sample["gt"])
                for sample in variant["samples"]
                if sample["name"] in samples_names
            )
            cursor.executemany(
                f"""INSERT INTO sample_has_variant VALUES (?,{variant_id},?)""", g
            )

        # Yield progression
        if variant_count % yield_every == 0:
            if total_variant_count:
                progress = variant_count / total_variant_count * 100

            yield progress, f"{variant_count} variants inserted."

    # Commit the transaction
    conn.commit()

    yield 97, f"{variant_count - errors} variant(s) has been inserted."

    # Create default selection (we need the number of variants for this)
    insert_selection(
        conn, "", name=cm.DEFAULT_SELECTION_NAME, count=variant_count - errors
    )


def insert_many_variants(conn, data, **kwargs):
    for _, _ in async_insert_many_variants(conn, data, kwargs):
        pass


## ================ Samples functions ==========================================


def create_table_samples(conn):
    """Create samples table

    :param conn: sqlite3.connect
    """
    cursor = conn.cursor()
    # sample_id is an alias on internal autoincremented 'rowid'
    cursor.execute(
        """CREATE TABLE samples (
        id INTEGER PRIMARY KEY ASC,
        name TEXT)"""
    )
    conn.commit()


def insert_sample(conn, name="no_name"):
    """Insert one sample in samples table (USED in TESTS)

    :param conn: sqlite3.connect
    :return: Last row id
    :rtype: <int>
    """
    cursor = conn.cursor()
    cursor.execute("""INSERT INTO samples (name) VALUES (?)""", [name])
    conn.commit()
    return cursor.lastrowid


def insert_many_samples(conn, samples: list):
    """Insert many samples at a time in samples table

    :param samples: List of samples names
        .. todo:: only names in this list ?
    :type samples: <list <str>>
    """
    cursor = conn.cursor()
    cursor.executemany(
        """INSERT INTO samples (name) VALUES (?)""", ((sample,) for sample in samples)
    )
    conn.commit()


def get_samples(conn):
    """"Get samples from sample table

    :param con: sqlite3.conn
    :return: Generator of dictionnaries with as sample fields as values.
        :Example: ({'id': <unique_id>, 'name': <sample_name>})
    :rtype: <generator <dict>>
    """
    conn.row_factory = sqlite3.Row
    return (dict(data) for data in conn.execute("""SELECT * FROM samples"""))


class Selection:
    """Binding object over selection which allows to do set operations on variants
    associated to it

    Attributes:
        - cls.conn: Class attribute for sqlite3 connection.
        - self.query: SQL query ready to be used in set operations.
        - self.mode: Define the way variants are compared to each other.
            - "variant" (default): chr,pos,ref,alt = variant.id
            - "site": chr,pos
        ..seealso:: from_selection_id()
    """

    # Class attribute, shared accross all instances
    conn = None

    def __init__(self, sql_query=None, mode="variant"):
        """Create a new selection object"""
        self.sql_query = sql_query
        # Define the way variants are compared
        self.mode = mode

    def __add__(self, other):
        sql_query = union_variants(self.sql_query, other.sql_query, mode=self.mode)
        return Selection(sql_query, self.mode)

    def __and__(self, other):
        sql_query = intersect_variants(self.sql_query, other.sql_query, mode=self.mode)
        return Selection(sql_query, self.mode)

    def __sub__(self, other):
        sql_query = subtract_variants(self.sql_query, other.sql_query, mode=self.mode)
        return Selection(sql_query, self.mode)

    def __repr__(self):
        return "<Selection>: " + self.sql_query

    def save(self, name):
        """Create the new selection in the database"""
        return create_selection_from_sql(
            Selection.conn, self.sql_query, name, from_selection=True
        )

    @classmethod
    def from_selection_id(cls, selection_id, mode="variant"):
        """Get new Selection object based on a sql query that defines it

        .. note:: Called from the UI. It is here that 'mode' is selected.

        :param selection_id: The id of the selection for which the object will be
            based.
        :key mode: Modifies the definition of the unicity of a variant.
            (optional: "variant" | "site"),(default: "variant").
            - variant: (chr, pos, ref, alt) is the primary key so we query only
            'selection_has_variant' for 'variant_id' field.
            - site: (chr, pos) modifies the default definition of the unicity
            of a variant.
            => joint on 'variants' table is mandatory here.
        :type selection_id: <int>
        :type mode: <str>
        :return: A new Selection object.
        :rtype: <Selection>
        """
        if selection_id == 1:
            # Get ids from the default selection
            sql_query = """SELECT id as variant_id FROM variants"""
        else:
            # A variant is defined as chr,pos,ref,alt
            # which is the primary key (unique)
            # So these columns are synonyms of 'variants.id' for the table 'variants'
            # or 'variant_id' for 'selection_has_variant' table.
            # No further joints are required here.
            sql_query = f"""SELECT variant_id
            FROM selection_has_variant sv
             WHERE sv.selection_id = {selection_id}"""

        return cls(sql_query, mode=mode)



# Standard imports
import vcf

# Custom imports
from .abstractreader import AbstractReader
from .annotationparser import VepParser, SnpEffParser
from cutevariant.commons import logger

LOGGER = logger()

VCF_TYPE_MAPPING = {"Float": "float", "Integer": "int", "Flag": "bool", "String": "str"}


class VcfReader(AbstractReader):
    """VCF parser to extract data from vcf file

    .. seealso:: AbstractReader class for more information.

    Attributes:
        annotation_parser (object): Support "VepParser()" and "SnpeffParser()"
    """

    def __init__(self, device, annotation_parser: str = None):
        """Construct a VCF Reader

        :param device: File device handler returned by open.
        :key annotation_parser (str): "vep" or "snpeff"
            This argument forces the reader to use a specific parser for
            the annotations. By default it's None: no parser will be used,
            annotations will not be taken into account.
        """
        super().__init__(device)

        vcf_reader = vcf.VCFReader(device)
        self.samples = vcf_reader.samples
        self.annotation_parser = None
        self._set_annotation_parser(annotation_parser)

    def get_fields(self):
        """Get full fields descriptions

        This function is called a first time before variants insertion.

        .. note:: Annotations fields are added here if they exist in the file.

        .. seealso:: parse_fields()

        :return: Tuple of fields.
            Each field is a dict with the following keys:
                name, category, description, type
            Some fields have an additional constraint key when they are destined
            to be a primary key in the database.
            Annotations fields are added here if they exist in the file.
            .. seealso parse_fields() for basic default fields.
        :rtype: <tuple <dict>>
        """
        LOGGER.debug("CsvReader::get_fields: called")
        if not self.fields:
            LOGGER.debug("CsvReader::get_fields: parse")
            fields = self.parse_fields()
            if self.annotation_parser:
                # If "ANN" is a field in the current VCF:
                # Remove and parse special annotations
                self.fields = tuple(self.annotation_parser.parse_fields(fields))
            else:
                self.fields = tuple(fields)
        return self.fields

    def get_variants(self):
        """Get variants as an iterable of dictionnaries

        "annotations" key is added here with the list of annotations if
        they exist in the file.

        .. seealso:: parse_variant()

        :return: Generator of full variants with "annotations" key.
        :rtype: <generator <dict>>
        """
        if self.annotation_parser:
            yield from self.annotation_parser.parse_variants(self.parse_variants())
        else:
            yield from self.parse_variants()

    def parse_variants(self):
        """Read file and parse variants

        .. note:: An estimation of the progression is made here by updating
            self.read_bytes attribute.

        :return: Generator of variants.
        :rtype: <generator <dict>>
        """
        # loop over record
        self.device.seek(0)
        vcf_reader = vcf.VCFReader(self.device)

        # TODO : ugly for testing progression .. see #60
        self.read_bytes = self._init_read_bytes(vcf_reader)

        for record in vcf_reader:

            self.read_bytes += self._get_record_size(record)

            # split row with multiple alt
            for index, alt in enumerate(record.ALT):
                # Remap some columns
                variant = {
                    "chr": record.CHROM,
                    "pos": record.POS,
                    "ref": record.REF,
                    "alt": str(alt),
                    "rsid": record.ID,  # Avoid id column duplication in DB
                    "qual": record.QUAL,
                    "filter": "",  # TODO ?,
                }

                # Parse info
                for name in record.INFO:
                    if isinstance(record.INFO[name], list):
                        variant[name.lower()] = ",".join(
                            [str(i) for i in record.INFO[name]]
                        )
                    else:
                        variant[name.lower()] = record.INFO[name]

                # parse sample
                if record.samples:
                    variant["samples"] = []
                    for sample in record.samples:
                        sample_data = {}
                        sample_data["name"] = sample.sample
                        sample_data["gt"] = (
                            -1 if sample.gt_type == None else sample.gt_type
                        )
                        variant["samples"].append(sample_data)

                yield variant

                #     # #PARSE Annotation
                #     # if category == "annotation": #=== PARSE Special Annotation ===
                #     #     # each variant can have multiple annotation. Create then many variants
                #     #     variant["annotation"] = []
                #     #     annotations = record.INFO["ANN"]
                #     #     for annotation in annotations:
                #     #         variant["annotation"].append(
                #     #             self.parser.parse_variant(annotation)
                #     #         )

    def parse_fields(self):
        """Extract fields informations from VCF fields

        .. note:: Fields used in PRIMARY KEYS have the constraint NOT NULL.
            By default, all other fields can have NULL values.

        :return: Generator of fields.
            Each field is a dict with the following keys:
                name, category, description, type
            Some fields have an additional constraint key when they are destined
            to be a primary key in the database.
        :rtype: <generator <dict>>
        """
        yield {
            "name": "chr",
            "category": "variants",
            "description": "chromosom",
            "type": "str",
            "constraint": "NOT NULL",
        }
        yield {
            "name": "pos",
            "category": "variants",
            "description": "position",
            "type": "int",
            "constraint": "NOT NULL",
        }
        yield {
            "name": "ref",
            "category": "variants",
            "description": "reference base",
            "type": "str",
            "constraint": "NOT NULL",
        }
        yield {
            "name": "alt",
            "category": "variants",
            "description": "alternative base",
            "type": "str",
            "constraint": "NOT NULL",
        }
        yield {
            "name": "rsid",
            "category": "variants",
            "description": "rsid",
            "type": "str",
        }
        yield {
            "name": "qual",
            "category": "variants",
            "description": "quality",
            "type": "int",
        }
        yield {
            "name": "filter",
            "category": "variants",
            "description": "filter",
            "type": "str",
        }

        # Reads VCF INFO
        self.device.seek(0)
        vcf_reader = vcf.VCFReader(self.device)

        #  Reads VCF info
        for key, info in vcf_reader.infos.items():

            # if key == "ANN": # Parse special annotation
            #     yield from self.parser.parse_fields(info.desc)
            # else:

            # Fix #75 : Avoid dot caracter in fields ...
            key = key.replace(".", "_")

            yield {
                "name": key.lower(),
                "category": "variants",
                "description": info.desc,
                "type": VCF_TYPE_MAPPING[info.type],
            }

        # Reads VCF FORMAT
        for key, info in vcf_reader.formats.items():
            yield {
                "name": key.lower(),
                "category": "samples",
                "description": info.desc,
                "type": VCF_TYPE_MAPPING[info.type],
            }

    def get_samples(self):
        """Return list of samples."""
        return self.samples

    # def _keep_unique_fields(self,fields):
    #     ''' return fields list with unique field name '''
    #     names = []
    #     for field in fields:
    #         if field["name"] not in names:
    #             names.append(field["name"])
    #             yield field

    # else:
    #     # Rename duplicate fields : field_1, field_2 etc ...
    #     field["name"]  = field["name"] +"_"+ str(names.count(field["name"])+1)
    #     yield field

    def _set_annotation_parser(self, parser: str):
        """Set the given annotation parser"""
        if parser == "vep":
            self.annotation_parser = VepParser()

        if parser == "snpeff":
            self.annotation_parser = SnpEffParser()

    def _get_record_size(self, rec):
        """Approximate record size in bytes"""
        # TODO : ugly .. For testing progression
        return len(
            str(rec.CHROM)
            + str(rec.POS)
            + str(rec.ID)
            + str(rec.REF)
            + str(rec.ALT)
            + str(rec.QUAL)
            + str(rec.FILTER)
            + str(rec.INFO)
            + str(rec.FORMAT)
            + str(rec.samples)
        )

    def _init_read_bytes(self, reader):
        """Init read bytes : It's the size in bytes of header data file"""
        return len("".join(reader._column_headers)) + len("".join(reader._header_lines))

    def __repr__(self):
        return f"VCF Reader using {type(self.annotation_parser).__name__}"



"""Exposes VqlEditor class used in the MainWindow to show SQL statements.

VqlEditor uses:
    - VqlSyntaxHighlighter as a syntax highlighter
    - VqlEdit for the support of autocompletion
"""
# Standard imports
from textx.exceptions import TextXSyntaxError

# Qt imports
from PySide2.QtCore import qApp, Qt, QRegularExpression, QStringListModel
from PySide2.QtWidgets import QTextEdit, QCompleter, QVBoxLayout, QLabel, QFrame
from PySide2.QtGui import (
    QSyntaxHighlighter,
    QFont,
    QPalette,
    QTextCharFormat,
    QTextCursor,
)

# Custom imports
from cutevariant.core import sql
from cutevariant.core.vql import VQLSyntaxError
from cutevariant.core import vql
from cutevariant.commons import MIN_COMPLETION_LETTERS, logger
from cutevariant.gui.ficon import FIcon
from cutevariant.gui import style
from .plugin import QueryPluginWidget

LOGGER = logger()


class VqlSyntaxHighlighter(QSyntaxHighlighter):
    """SQL Syntax highlighter for VqlEditor"""

    sql_keywords = (
        "SELECT",
        "FROM",
        "WHERE",
        "AS",
        "AND",
        "OR",
        "NOT",
        "ALL",
        "ANY",
        "BETWEEN",
        "EXISTS",
        "IN",
        "LIKE",
        "SOME",
        "ASC",
        "DESC",
        "LIMIT",
        "DISTINCT",
        "GROUP BY",
        "HAVING",
        "ORDER BY",
        "IS",
        "NOT",
        "NULL",
    )

    def __init__(self, document=None):
        super().__init__(document)

        palette = qApp.palette("QTextEdit")

        # SQL Syntax highlighter rules
        # dict: pattern, font, color, minimal (not greedy)
        #  TODO : What about dark mode ?
        self.highlighting_patterns = [
            {
                # Keywords
                # \b allows to perform a "whole words only"
                "pattern": "|".join(
                    (
                        "\\b%s\\b" % keyword
                        for keyword in VqlSyntaxHighlighter.sql_keywords
                    )
                ),
                "font": QFont.Bold,
                "color": palette.color(QPalette.Highlight),  # default: Qt.darkBlue
                "case_insensitive": True,
            },
            {
                # Strings simple quotes '...'
                "pattern": r"\'.*\'",
                "color": Qt.red,
                "minimal": True,  # Need to stop match as soon as possible
            },
            {
                # Strings double quotes: "..."
                "pattern": r"\".*\"",
                "color": Qt.red,
                "minimal": True,  # Need to stop match as soon as possible
            },
            {
                # Numbers
                "pattern": r"\\b[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?\\b",
                "color": Qt.darkGreen,
            },
            {
                # Comments
                "pattern": r"--[^\n]*",
                "color": Qt.gray,
            },
        ]

        self.highlighting_rules = list()

        for pattern in self.highlighting_patterns:

            t_format = QTextCharFormat()
            font = pattern.get("font", None)
            if font:
                t_format.setFontWeight(font)

            color = pattern.get("color", None)
            if color:
                t_format.setForeground(color)

            regex = QRegularExpression(pattern["pattern"])
            if pattern.get("minimal", False):
                # The greediness of the quantifiers is inverted: *, +, ?, {m,n}, etc.
                # become lazy, while their lazy versions (*?, +?, ??, {m,n}?, etc.)
                # become greedy.
                # https://doc.qt.io/Qt-5/qregularexpression.html#setPatternOptions
                regex.setPatternOptions(QRegularExpression.InvertedGreedinessOption)

            if pattern.get("case_insensitive", False):
                # NOTE: Deletes previous pattern options
                # Not serious in practice, this only concerns the keywords
                regex.setPatternOptions(QRegularExpression.CaseInsensitiveOption)

            self.highlighting_rules.append((regex, t_format))

    def highlightBlock(self, text):
        """Overrided"""
        for regex, t_format in self.highlighting_rules:
            # Ugly iterator => not iterable in Python...
            matchIterator = regex.globalMatch(text)

            while matchIterator.hasNext():
                match = matchIterator.next()
                self.setFormat(match.capturedStart(), match.capturedLength(), t_format)


class VqlEditor(QueryPluginWidget):
    """Exposed class to manage VQL/SQL queries from the mainwindow"""

    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("Columns"))

        # Syntax highlighter and autocompletion
        self.text_edit = VqlEdit()
        self.log_edit = QLabel()
        self.highlighter = VqlSyntaxHighlighter(self.text_edit.document())

        self.log_edit.setMinimumHeight(40)
        # self.log_edit.setAlignment(Qt.AlignLeft|Qt.AlignVCenter)
        self.log_edit.setStyleSheet(
            "QWidget{{background-color:'{}'; color:'{}'}}".format(
                style.WARNING_BACKGROUND_COLOR, style.WARNING_TEXT_COLOR
            )
        )
        self.log_edit.hide()

        self.log_edit.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
        main_layout = QVBoxLayout()
        main_layout.addWidget(self.text_edit)
        main_layout.addWidget(self.log_edit)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)
        self.setLayout(main_layout)
        self._query = None

    def on_init_query(self):
        """Overrided"""
        self.text_edit.setCompleter(self.create_completer())
        self.on_change_query()

    def on_change_query(self):
        """Overrided"""
        self.text_edit.blockSignals(True)
        self.text_edit.setPlainText(self.query.to_vql())
        self.text_edit.blockSignals(False)

    def create_completer(self):
        """Create Completer with his model"""
        model = QStringListModel()
        completer = QCompleter()
        # Fill the model with the SQL keywords and database fields
        fields = [i["name"] for i in sql.get_fields(self._query.conn)]
        fields.extend(VqlSyntaxHighlighter.sql_keywords)
        model.setStringList(fields)
        completer.setModel(model)
        return completer

    def check_vql(self):
        """Check VQL statement; return True if OK, False when an error occurs

        .. note:: This function also sets the error message to the bottom of the view.
        """
        try:
            self.log_edit.hide()
            tuple(vql.execute_vql(self.text_edit.toPlainText()))

        except TextXSyntaxError as e:
            # Available attributes: e.message, e.line, e.col
            self.set_message("TextXSyntaxError: %s, col: %d" % (e.message, e.col))
            return False

        except VQLSyntaxError as e:
            # Show the error message on the ui
            self.set_message(
                self.tr("VQLSyntaxError: '%s' at position %s") % (e.message, e.col)
            )
            return False

        return True

    def run_vql(self):
        """Execute VQL query"""
        # self.query_changed.emit()

        if not self.check_vql():
            return

        for cmd in vql.execute_vql(self.text_edit.toPlainText()):

            #  If command is a select kind
            if cmd["cmd"] == "select_cmd":
                self.query.columns = cmd["columns"]  # columns from variant table
                self.query.selection = cmd["source"]  # name of the variant set
                self.query.filter = cmd.get(
                    "filter", dict()
                )  # filter as raw text; dict if no filter
                self.query_changed.emit()

            if cmd["cmd"] == "create_cmd":
                #  TODO create selection
                pass

    def set_message(self, message: str):
        """Show message error at the bottom of the view"""

        if self.log_edit.isHidden():
            self.log_edit.show()

        icon_64 = FIcon(0xF5D6, style.WARNING_TEXT_COLOR).to_base64(18, 18)

        self.log_edit.setText(
            """
            <div height=100%>
            <img src="data:image/png;base64,{}" align="left"/>
             <span>  {} </span>
            </div>""".format(
                icon_64, message
            )
        )


class VqlEdit(QTextEdit):
    """Custom class inheriting from QTextEdit, used by VqlEditor"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.completer = None
        # Character that triggers the full autocompletion list
        self.completer_joker = "!"

    def setCompleter(self, completer: QCompleter):
        """Register and init the given QCompleter to the QTextEdit"""
        if self.completer:
            self.completer.activated.disconnect()

        self.completer = completer
        self.completer.setWidget(self)
        self.completer.setCompletionMode(QCompleter.PopupCompletion)
        self.completer.setCaseSensitivity(Qt.CaseInsensitive)
        self.completer.activated.connect(self.insertCompletion)

    def keyPressEvent(self, event):
        """Overrided"""
        # Ignore some key events so the completer can handle them.
        if self.completer and self.completer.popup().isVisible():
            if event.key() in (
                Qt.Key_Enter,
                Qt.Key_Return,
                Qt.Key_Escape,
                Qt.Key_Tab,
                Qt.Key_Backtab,
            ):
                event.ignore()
                # Let the completer do default behavior
                return

        # Accept other key events
        super().keyPressEvent(event)

        # Ignore modifiers
        found_ignored_modifier = event.modifiers() in (
            Qt.ControlModifier,
            Qt.ShiftModifier,
            Qt.AltModifier,
        )

        # LOGGER.debug("keyPressEvent:: event text: %s", event.text())

        # Dismiss ingored modifiers without text
        if not self.completer or (found_ignored_modifier and not event.text()):
            LOGGER.debug("keyPressEvent:: ignored modifier")
            return

        has_modifier = event.modifiers() != Qt.NoModifier and not found_ignored_modifier

        end_of_word = "~@#$%^&*()_+{}|:\"<>?,./;'[]\\-="
        completion_prefix = self.textUnderCursor()
        completer = self.completer

        # LOGGER.debug("keyPressEvent:: has_modifier: %s", has_modifier)
        # LOGGER.debug("keyPressEvent:: completion_prefix: %s", completion_prefix)

        # Hide on alone modifier, empty text, short text, end of word
        if self.completer_joker not in event.text() and (
            has_modifier
            or not event.text()
            or len(completion_prefix) < MIN_COMPLETION_LETTERS
            or event.text()[-1] in end_of_word
        ):
            completer.popup().hide()
            # LOGGER.debug("keyPressEvent:: Hide completer popup")
            return

        # Select proposed word
        if completion_prefix != completer.completionPrefix():
            if self.completer_joker in event.text():
                # Joker is found: display the full list
                completer.setCompletionPrefix("")
            else:
                completer.setCompletionPrefix(completion_prefix)
            completer.popup().setCurrentIndex(completer.completionModel().index(0, 0))

        # Show popup
        cursor_rect = self.cursorRect()
        cursor_rect.setWidth(
            completer.popup().sizeHintForColumn(0)
            + completer.popup().verticalScrollBar().sizeHint().width()
        )
        completer.complete(cursor_rect)

    def focusInEvent(self, event):
        """Overrided: Event handler used to receive keyboard focus events for the widget"""
        if self.completer:
            self.completer.setWidget(self)

        super().focusInEvent(event)

    def insertCompletion(self, completion: str):
        """Complete the word using the given completion string

        .. note:: Called after user validation in the popup

        :param completion: Word proposed by the autocompletion.
        :type completion: <str>
        """
        # Ensure that the completer's widget is the current one
        if self.completer.widget() != self:
            return

        tc = self.textCursor()
        # Get number of characters that must be inserted
        # The last nb_extra characters from the right will be inserted by tc
        nb_extra = len(completion) - len(self.completer.completionPrefix())
        # Do not replace anything if the word is already completed
        # => avoid word duplication when the cursor is not moved (nb_extra = 0)
        if nb_extra == 0:
            return

        tc.movePosition(QTextCursor.Left)  # left one character.
        tc.movePosition(QTextCursor.EndOfWord)  # end of the current word.
        # Get a substring that contains the nb_extra rightmost characters
        # of the string; and insert the extra characters to complete the word.
        tc.insertText(completion[-nb_extra:])

        # Erase the joker
        tc.movePosition(QTextCursor.StartOfWord)
        current_char = self.toPlainText()[tc.position() - 1]
        if current_char == self.completer_joker:
            tc.deletePreviousChar()

        # Move cursor after the word
        tc.movePosition(QTextCursor.EndOfWord)
        self.setTextCursor(tc)

    def textUnderCursor(self):
        """Select a word under the cursor and return it

        :return: The text/fragment of word under the cursor.
        :rtype: <QTextCursor>
        """
        tc = self.textCursor()
        tc.select(QTextCursor.WordUnderCursor)
        return tc.selectedText()



"""Plugin to show base changes chart based on the current query"""
# Standard imports
import itertools as it
from collections import Counter, defaultdict
from copy import copy, deepcopy
from logging import DEBUG

# Qt imports
from PySide2.QtWidgets import QVBoxLayout
from PySide2.QtCore import Qt
from PySide2.QtGui import QPainter
from PySide2.QtCharts import QtCharts as charts

# Custom imports
from .plugin import QueryPluginWidget
from cutevariant.commons import logger, DEFAULT_SELECTION_NAME

LOGGER = logger()


class ChartQueryWidget(QueryPluginWidget):
    """Plugin to show charts based on the current query"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle(self.tr("Charts"))

        self.view = charts.QChartView()

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.view)

        self.setLayout(layout)
        self._query = None
        self.previous_filter = ""
        self.previous_selection = ""

    def on_init_query(self):
        """Overrided from AbstractQueryWidget"""
        pass

    def on_change_query(self):
        """Build a QBarSeries which displays base changes for the current variants
        in the query

        .. note:: Called when the VQL query is updated.

        .. TODO:: not optimal, this is running many time when other queryWidget changed
            Iterate over ALL variants... can be slow. Do I need asynchrone?

        """
        # BarSet: List of cccurrences of 1 alternative base in all snps for 1 reference base
        # => 4 values for each 4 bases of reference
        # => 4 barsets

        # QBarSet *set0 = new QBarSet("Jane");
        # *set0 << 1 << 2 << 3 << 4 << 5 << 6;
        # QBarSeries *series = new QBarSeries();
        # series->append(set0);

        # print("cols", self.query.columns)
        # print("filter", self.query.filter)
        # print("groub_by", self.query.group_by)
        # print("selec", self.query.selection)

        # Do not redo the chart on if data is not changed...
        if (
            self.previous_filter == self.query.filter
            and self.previous_selection == self.query.selection
        ):
            LOGGER.debug(
                "ChartQueryWidget:on_change_query: Query filter didn't change => Do nothing!"
            )
            return

        self.previous_filter = self.query.filter
        self.previous_selection = self.query.selection

        # We don't want:
        # SELECT chr,pos,ref,alt,COUNT(variants.id) as 'children'
        # FROM variants
        # LEFT JOIN annotations ON annotations.variant_id = variants.id
        # WHERE ref IN ('A', 'T', 'G', C) AND alt IN ('A', 'T', 'G', C)
        # GROUP BY chr,pos,ref,alt

        # We just want:
        # SELECT re, alt, COUNT(*) FROM variants
        # WHERE ref IN ('A', 'T', 'G', C) AND alt IN ('A', 'T', 'G', C)
        # GROUP by ref, alt

        # AND we HAVE TO take care of current filters and selections;
        # so we will reset all fields except this ones; filter attribute
        # will be edited in place to add our new filters.

        query = copy(self.query)
        query.group_by = ("ref", "alt")
        query.order_by = None
        query.columns = ["ref", "alt", "COUNT(*)"]
        query.filter = deepcopy(self.query.filter)
        if query.selection == DEFAULT_SELECTION_NAME:
            # Explicitly query all variants
            query.selection = None
        # Example of filters:
        # filter: {'AND': [{'field': 'ref', 'operator': '=', 'value': 'G'},
        #                  {'field': 'ref', 'operator': 'IN', 'value': "('A', 'T', 'G', C)"}]}
        # filter: {'AND': [{'field': 'ref', 'operator': 'IN', 'value': "('A', 'T', 'G', 'C')"},
        #                  {'field': 'alt', 'operator': 'IN', 'value': "('A', 'T', 'G', 'C')"}]}
        # We want to add our filter in the top of filters:
        # - If there is no filters:
        # We add a new AND filter with our filters in its list.
        # - If AND operator is on TOP:
        # Just add our filters to the list of this operator.
        # - If OR operator is on TOP:
        # We have to encapsulate the whole tree of filter into a AND one;
        # so we add our filters to this new list just after the old OR filter.
        ATGC_filter = [
            {"field": "ref", "operator": "IN", "value": "('A', 'T', 'G', 'C')"},
            {"field": "alt", "operator": "IN", "value": "('A', 'T', 'G', 'C')"},
        ]

        # Avoid touching to original data since we don't do deepcopy (SQLite is not pickable)
        query.filter = deepcopy(query.filter)
        if not query.filter:
            # no filters: new AND filter
            query.filter["AND"] = ATGC_filter
        elif "AND" in query.filter:
            # AND operator on TOP: add our filters to it
            query.filter["AND"].extend(ATGC_filter)
        else:
            # OR operator on TOP: encapsulate the old filter into a AND one
            # and add our filters to this new list just after the old OR filter.
            query.filter = {"AND": [query.filter]}  # Old OR filter
            query.filter["AND"].extend(ATGC_filter)

        ## Data formatting
        # Raw variants:
        # {'ref': 'G', 'alt': 'T', 'COUNT(*)': 1}
        # Build  dict like:
        # {'G': Counter({'T': 3, 'A': 1}), 'C': ...}

        data = defaultdict(Counter)

        if LOGGER.getEffectiveLevel() == DEBUG:
            LOGGER.debug("ChartQueryWidget:on_change_query:: Custom query built:")
            import time

            start = time.time()

        # After the auto-parsing of filters by query.sql(), add manually:
        # - COUNT() function to columns
        # - GROUP BY command to query
        # We use RAW factory as tuples
        for ref, alt, value in query.conn.execute(
            query.sql(do_not_add_default_things=True)
        ):
            data[ref][alt] = value

        if LOGGER.getEffectiveLevel() == DEBUG:
            end_query = time.time()
            LOGGER.debug(
                "ChartQueryWidget:on_change_query:: number %s, query time %s",
                len(data),
                end_query - start,
            )

        ## Create QBarSets
        references = ("A", "T", "G", "C")
        # alt bases as keys; barsets as values
        barsets = dict()

        # We iterate on data always in the same order (see references variable)
        # AA, AT, AG, AC, TA, ...
        max_value = 0
        for ref, alt in it.product(references, repeat=2):

            if alt not in barsets:
                # Init the barset for this alt base
                barset = charts.QBarSet(alt)
                barsets[alt] = barset
            else:
                barset = barsets[alt]

            # Get value and append it to the current barset
            # As it is a default dict with Counters; Missing bases have the value 0
            value = data[ref][alt]
            max_value = value if value > max_value else max_value
            # print("Append to %s, %s" % (alt, value))
            barset.append(value)

        ## Build Chart
        # Add barsets to QBarSeries
        bar_series = charts.QBarSeries()
        bar_series.append(tuple(barsets.values()))  # only takes list or tuple

        # Init chart, set title, turn on animations
        chart = charts.QChart()
        chart.addSeries(bar_series)
        chart.setTitle(self.tr("SNPs Base Changes"))
        chart.setAnimationOptions(charts.QChart.SeriesAnimations)

        # Add categories for x-axis
        axisX = charts.QBarCategoryAxis()
        axisX.append(references)
        chart.addAxis(axisX, Qt.AlignBottom)
        bar_series.attachAxis(axisX)

        # Add y-axis with the maximum value encountered
        axisY = charts.QValueAxis()
        axisY.setRange(0, max_value + 1)
        chart.addAxis(axisY, Qt.AlignLeft)
        bar_series.attachAxis(axisY)

        # Add legend
        chart.legend().setVisible(True)
        chart.legend().setAlignment(Qt.AlignBottom)

        # Add the chart to the view, turn on the antialiasing
        self.view.setChart(chart)
        self.view.setRenderHint(QPainter.Antialiasing)
